import os
from abc import ABC, abstractmethod
from typing import List, Set, Tuple, Union

from metabolights_utils.isatab.default.parser.investigation_parser import (
    parse_investigation_from_fs,
)
from metabolights_utils.isatab.default.parser.isa_table_parser import (
    assay_file_expected_patterns,
    parse_isa_table_sheet_from_fs,
    samples_file_expected_patterns,
)
from metabolights_utils.models.common import GenericMessage, GenericMessageType
from metabolights_utils.models.isa.assay_file import AssayFile
from metabolights_utils.models.isa.assignment_file import AssignmentFile
from metabolights_utils.models.isa.common import (
    IsaTableFile,
    OntologyItem,
    OrganismAndOrganismPartPair,
)
from metabolights_utils.models.isa.investigation_file import Investigation
from metabolights_utils.models.isa.samples_file import SamplesFile
from metabolights_utils.models.metabolights.model import (
    MetabolightsStudyModel,
    StudyDBMetadata,
    StudyFolderMetadata,
)
from metabolights_utils.models.parser.common import ParserMessage
from metabolights_utils.models.parser.enums import ParserMessageType
from metabolights_utils.provider.utils import (
    find_assay_technique,
    get_unique_file_extensions,
)


class AbstractDbMetadataCollector(ABC):

    def __init__(self):
        pass

    @abstractmethod
    def get_study_metadata_from_db(
        self, study_id: str, connection
    ) -> Tuple[Union[StudyDBMetadata, None], List[GenericMessage]]:
        """_summary_

        Args:
            study_id (str): MetaboLights study accession number.
            connection (Any): remote db connection to fetch study metadata.


        Returns:
            StudyDBMetadata: study metadata object that contains all the study db metadata fields.
            messages: List[GenericMessage]: list of messages to be populated with any errors or warnings.
        """
        pass


class AbstractFolderMetadataCollector(ABC):
    def __init__(self):
        pass

    @abstractmethod
    def get_folder_metadata(
        self,
        study_path,
        calculate_data_folder_size: bool = False,
        calculate_metadata_size: bool = False,
    ) -> Tuple[Union[StudyFolderMetadata, None], List[GenericMessage]]:
        """_summary_

        Args:
            study_path (_type_): MetaboLights study accession number.
            calculate_data_folder_size (bool, optional): calculate study FILES folder size. Defaults to False.
            calculate_metadata_size (bool, optional): calculate size of study ISA metadata files
                even if they are referenced or not. Defaults to False.
                If calculate_data_folder_size and calculate_metadata_size are False, folder size will be set to -1.

        Returns:
            StudyFolderMetadata: study folder metadata object that contains all descriptors of the study folders and files.
                .d, .raw and .m folders are included but not their content. If there is an error, None is returned.
            messages: List[GenericMessage]: list of messages to be populated with any errors or warnings.
        """
        pass


class MetabolightsStudyProvider(object):

    def __init__(
        self,
        db_metadata_collector: Union[None, AbstractDbMetadataCollector] = None,
        folder_metadata_collector: Union[None, AbstractFolderMetadataCollector] = None,
    ) -> None:
        self.db_metadata_collector = db_metadata_collector
        self.folder_metadata_collector = folder_metadata_collector

    def _add_parse_messages(
        self,
        model: MetabolightsStudyModel,
        fileName: str,
        messages: List[ParserMessage],
    ):
        if fileName not in model.parser_messages:
            model.parser_messages[fileName] = []
        model.parser_messages[fileName].extend(self.filter_messages(messages))

    def set_organisms(self, samples_file: SamplesFile, isa_table: SamplesFile):
        pairs = set()
        columns = isa_table.table.columns
        data = isa_table.table.data
        characteristics_columns = [
            "Characteristics[Organism]",
            "Characteristics[Organism part]",
            "Characteristics[Variant]",
            "Characteristics[Sample type]",
        ]
        characteristics_set = {x: set() for x in characteristics_columns}
        source_ref_column_prefix = "Term Source REF"
        accession_column_prefix = "Term Accession Number"
        characteristics_column_indices = {
            x: columns.index(x) for x in characteristics_columns if x in columns
        }
        if not samples_file.table.data or not samples_file.table.columns:
            return

        for i in range(len(data[columns[0]])):
            characteristics: Tuple[
                OntologyItem, OntologyItem, OntologyItem, OntologyItem
            ] = (OntologyItem(), OntologyItem(), OntologyItem(), OntologyItem())
            for idx, column_name in enumerate(characteristics_columns):
                if column_name in columns:
                    characteristics[idx].term = data[column_name][i]
                index = characteristics_column_indices[column_name]
                source_ref = columns[index + 1] if index + 1 < len(columns) else ""

                if source_ref.startswith(source_ref_column_prefix):
                    characteristics[idx].term_source_ref = data[columns[index + 1]][i]
                accession = columns[index + 2] if index + 2 < len(columns) else ""

                if accession.startswith(accession_column_prefix):
                    characteristics[idx].term_accession_number = data[
                        columns[index + 2]
                    ][i]
                if characteristics[idx].term and characteristics[idx].term.strip():
                    characteristics_set[column_name].add(characteristics[idx])
            if characteristics[0].term and characteristics[0].term.strip():
                pairs.add(
                    OrganismAndOrganismPartPair(
                        organism=characteristics[0],
                        organismPart=characteristics[1],
                        variant=characteristics[2],
                        sampleType=characteristics[3],
                    )
                )
        samples_file.organisms = list(characteristics_set[characteristics_columns[0]])
        samples_file.organism_parts = list(
            characteristics_set[characteristics_columns[1]]
        )
        samples_file.variants = list(characteristics_set[characteristics_columns[2]])
        samples_file.sample_types = list(
            characteristics_set[characteristics_columns[3]]
        )
        samples_file.organism_and_organism_part_pairs = list(pairs)

    def get_phase1_input_data(
        self,
        study_id: str,
        folder: Union[str, None] = None,
        connection=None,
    ) -> MetabolightsStudyModel:
        model: MetabolightsStudyModel = MetabolightsStudyModel()

        self.update_investigation_file(model, folder)
        if self.db_metadata_collector and connection:
            self.update_study_db_metadata(
                model,
                self.db_metadata_collector,
                study_id=study_id,
                connection=connection,
            )
        else:
            message = GenericMessage(
                type=GenericMessageType.WARNING,
                short=f"Database metadata collection is skipped for {study_id}",
                detail=f"Database metadata collection is skipped for {study_id}",
            )
            model.db_reader_messages.append(message)

        if not model.investigation.studies:
            return model

        raw_files = set()
        derived_files = set()
        assignment_files = set()
        folders_in_hierarchy = set()
        investigation = model.investigation
        for study_item in investigation.studies:
            file_path = os.path.join(folder, study_item.file_name)
            samples_isa_table, messages = parse_isa_table_sheet_from_fs(
                file_path,
                samples_file_expected_patterns,
                limit=0,
                fix_unicode_exceptions=True,
            )
            self._add_parse_messages(model, study_item.file_name, messages)
            samples_file = SamplesFile()
            samples_file.file_path = study_item.file_name
            samples_file.table = samples_isa_table.table
            samples_file.sha256_hash = samples_isa_table.sha256_hash
            selected_column_names = set()

            selected_column_names = samples_isa_table.table.columns

            sample_names, message_list = parse_isa_table_sheet_from_fs(
                file_path,
                samples_file_expected_patterns,
                selected_columns=list(selected_column_names),
                fix_unicode_exceptions=True,
            )
            self._add_parse_messages(model, samples_file.file_path, message_list)

            samples_file.table.total_row_count = 0
            unique_sample_names = set()
            if sample_names.table.data and "Sample Name" in sample_names.table.data:
                samples_row = len(sample_names.table.data["Sample Name"])
                samples_file.table.total_row_count = samples_row

                unique_sample_names = {
                    f.strip() for f in sample_names.table.data["Sample Name"]
                }
            unique_sample_names.discard("")
            unique_sample_names.discard(None)
            samples_file.sample_names = list(unique_sample_names)
            self.set_organisms(samples_file, sample_names)
            model.samples[study_item.file_name] = samples_file
            for assay_item in study_item.study_assays.assays:
                file_path = os.path.join(folder, assay_item.file_name)
                assay_isa_table_sheet, messages = parse_isa_table_sheet_from_fs(
                    file_path,
                    assay_file_expected_patterns,
                    limit=0,
                    fix_unicode_exceptions=True,
                )
                assay_isa_table: IsaTableFile = assay_isa_table_sheet
                assay_file: AssayFile = AssayFile()
                assay_file.file_path = assay_item.file_name
                assay_file.table = assay_isa_table.table
                assay_file.sha256_hash = assay_isa_table.sha256_hash
                self._add_parse_messages(model, assay_item.file_name, messages)
                model.assays[assay_item.file_name] = assay_file

                selected_column_names = set()
                for column in assay_isa_table.table.columns:
                    if (
                        column.startswith("Raw Spectral Data File")
                        or column.startswith("Acquisition Parameter Data File")
                        or column.startswith("Free Induction Decay Data File")
                    ):
                        selected_column_names.add(column)
                    elif column.startswith("Derived Spectral Data File"):
                        selected_column_names.add(column)
                    elif column.startswith("Metabolite Assignment File"):
                        selected_column_names.add(column)
                    elif column == "Sample Name":
                        selected_column_names.add(column)
                    elif column == "MS Assay Name" or column == "NMR Assay Name":
                        selected_column_names.add(column)
                    elif column == "Parameter Value[Column type]":
                        selected_column_names.add(column)

                selected_column_names_list = list(selected_column_names)
                assay_file_subset_sheet, messages_list = parse_isa_table_sheet_from_fs(
                    file_path,
                    assay_file_expected_patterns,
                    selected_columns=selected_column_names_list,
                    fix_unicode_exceptions=True,
                )
                assay_file_subset: IsaTableFile = assay_file_subset_sheet
                self._add_parse_messages(model, assay_item.file_name, messages_list)
                if selected_column_names:
                    assay_file.table.total_row_count = 0
                    if assay_file_subset.table.data:
                        assay_file.table.total_row_count = len(
                            assay_file_subset.table.data[selected_column_names_list[0]]
                        )
                else:
                    column_names = assay_file.table.columns
                    if len(column_names) > 0:
                        (
                            first_column_of_table,
                            messages_list,
                        ) = parse_isa_table_sheet_from_fs(
                            file_path,
                            assay_file_expected_patterns,
                            selected_columns=[column_names[0]],
                            fix_unicode_exceptions=True,
                        )
                        first_column: IsaTableFile = first_column_of_table
                        self._add_parse_messages(
                            model, assay_item.file_name, messages_list
                        )
                        assay_file.table.total_row_count = 0
                        if first_column.table.data:
                            assay_file.table.total_row_count = len(
                                first_column.table.data[first_column.table.columns[0]]
                            )
                    else:
                        assay_file.table.total_row_count = 0
                    assay_file.table.row_count = 0
                assay_raw_files: Set = set()
                assay_raw_file_extensions: Set = set()
                assay_derived_files: Set = set()
                assay_derived_file_extensions = set()

                for column in assay_file_subset.table.columns:
                    file_set = {f.strip() for f in assay_file_subset.table.data[column]}
                    file_set.discard("")
                    file_set.discard(None)
                    file_list = list(file_set)
                    if (
                        column.startswith("Raw Spectral Data File")
                        or column.startswith("Acquisition Parameter Data File")
                        or column.startswith("Free Induction Decay Data File")
                    ):
                        raw_files.update(file_set)
                        assay_raw_files.update(file_list)
                        assay_raw_file_extensions.update(
                            get_unique_file_extensions(file_list)
                        )
                    elif column.startswith("Derived Spectral Data File"):
                        derived_files.update(file_set)
                        assay_derived_files.update(file_list)
                        assay_derived_file_extensions.update(
                            get_unique_file_extensions(file_list)
                        )
                    elif column.startswith("Metabolite Assignment File"):
                        assignment_files.update(file_set)
                        assay_file.referenced_assignment_files = file_list
                    elif column == "Sample Name":
                        assay_file.sample_names = file_list
                    elif column == "MS Assay Name" or column == "NMR Assay Name":
                        assay_file.assay_names = file_list
                assay_file.assay_technique = find_assay_technique(
                    model, assay_file, assay_file_subset
                )

                assay_file.referenced_raw_files = list(assay_raw_files)
                assay_file.referenced_derived_files = list(assay_derived_files)
                assay_file.referenced_raw_file_extensions = list(
                    assay_raw_file_extensions
                )
                assay_file.referenced_derived_file_extensions = list(
                    assay_derived_file_extensions
                )

        for set_item in (raw_files, derived_files, assignment_files):
            for item in set_item:
                if set_item != assignment_files:
                    folders_in_hierarchy.add(os.path.dirname(item))

        model.referenced_assignment_files.extend(list(assignment_files))
        model.referenced_raw_files.extend(list(raw_files))
        model.referenced_derived_files.extend(list(derived_files))
        model.folders_in_hierarchy.extend(list(folders_in_hierarchy))

        for assignment_file in assignment_files:
            absolute_path = os.path.join(folder, assignment_file)
            isa_table_sheet, messages = parse_isa_table_sheet_from_fs(
                absolute_path, limit=0, fix_unicode_exceptions=True
            )
            metabolite_assignment_isa_table: IsaTableFile = isa_table_sheet
            self._add_parse_messages(model, assignment_file, messages)
            file: AssignmentFile = AssignmentFile()
            file.file_path = assignment_file
            file.table = metabolite_assignment_isa_table.table
            file.sha256_hash = metabolite_assignment_isa_table.sha256_hash
            model.metabolite_assignments[assignment_file] = file
            column_names = file.table.columns

            selected_column_names = set()
            for column in file.table.columns:
                if (
                    column == "database_identifier"
                    or column == "metabolite_identification"
                ):
                    selected_column_names.add(column)

            if len(selected_column_names) > 0:
                selected_isa_table_data, messages_list = parse_isa_table_sheet_from_fs(
                    absolute_path,
                    selected_columns=list(selected_column_names),
                    fix_unicode_exceptions=True,
                )
                selected_column_values: IsaTableFile = selected_isa_table_data
                self._add_parse_messages(model, assignment_file, messages_list)

                file.table.total_row_count = 0
                metabolite_identifications = set()
                metabolite_assignments = {}
                if selected_column_values.table.data:
                    file.table.total_row_count = len(
                        selected_column_values.table.data[
                            selected_column_values.table.columns[0]
                        ]
                    )
                    for i in range(file.table.total_row_count):
                        database_identifier = ""
                        metabolite_identification = ""
                        for column in selected_column_names:
                            if column == "database_identifier":
                                database_identifier = selected_column_values.table.data[
                                    column
                                ][i]
                            elif column == "metabolite_identification":
                                metabolite_identification = (
                                    selected_column_values.table.data[column][i]
                                )
                        if metabolite_identification:
                            metabolite_identifications.add(metabolite_identification)
                            if database_identifier:
                                metabolite_assignments[metabolite_identification] = (
                                    database_identifier
                                )
                    file.identified_metabolite_names = list(metabolite_identifications)
                    file.metabolite_assignments = metabolite_assignments
                    file.number_of_assigned_rows = len(file.metabolite_assignments)
                    file.number_of_unassigned_rows = (
                        file.table.total_row_count - file.number_of_assigned_rows
                    )
                    file.number_of_rows = file.table.total_row_count
            else:
                file.table.total_row_count = 0
            file.table.row_count = 0

        for assay_file_name in model.assays:
            assay_file_item = model.assays[assay_file_name]
            for assignement_file_name in assay_file_item.referenced_assignment_files:
                model.metabolite_assignments[assignement_file_name].assay_technique = (
                    assay_file_item.assay_technique
                )
        return model

    def get_sample_file_input(
        self,
        study_id: str,
        folder: str,
        connection,
        samples_sheet_offset: Union[int, None] = None,
        samples_sheet_limit: Union[int, None] = None,
        model: MetabolightsStudyModel = None,
    ) -> MetabolightsStudyModel:
        if not model:
            model = self.get_phase1_input_data(study_id, folder, connection)

        for study_item in model.investigation.studies:
            file_path = os.path.join(folder, study_item.file_name)
            samples_isa_table_sheet, messages = parse_isa_table_sheet_from_fs(
                file_path,
                samples_file_expected_patterns,
                offset=samples_sheet_offset,
                limit=samples_sheet_limit,
                fix_unicode_exceptions=True,
            )
            samples_isa_table: IsaTableFile = samples_isa_table_sheet
            model.samples[study_item.file_name].table = samples_isa_table.table
            model.samples[study_item.file_name].sha256_hash = (
                samples_isa_table.sha256_hash
            )
            table = samples_isa_table.table
            table.row_count = 0
            if table.data:
                table.row_count = len(table.data[samples_isa_table.table.columns[0]])

            model.parser_messages[study_item.file_name].extend(
                self.filter_messages(messages)
            )
            model.has_sample_table_data = True
        return model

    def get_phase2_input_data(
        self,
        study_id: str,
        folder: Union[None, str],
        connection=None,
        samples_sheet_offset: Union[int, None] = None,
        samples_sheet_limit: Union[int, None] = None,
        assay_sheet_offset: Union[int, None] = None,
        assay_sheet_limit: Union[int, None] = None,
        model: MetabolightsStudyModel = None,
    ) -> MetabolightsStudyModel:
        if not model:
            model = self.get_phase1_input_data(study_id, folder, connection)

        for study_item in model.investigation.studies:
            file_path = os.path.join(folder, study_item.file_name)
            samples_isa_table_sheet, messages = parse_isa_table_sheet_from_fs(
                file_path,
                samples_file_expected_patterns,
                offset=samples_sheet_offset,
                limit=samples_sheet_limit,
                fix_unicode_exceptions=True,
            )
            samples_isa_table: IsaTableFile = samples_isa_table_sheet
            model.samples[study_item.file_name].table = samples_isa_table.table
            model.samples[study_item.file_name].sha256_hash = (
                samples_isa_table.sha256_hash
            )

            if samples_isa_table.table.data:
                model.samples[study_item.file_name].table.row_count = len(
                    samples_isa_table.table.data[samples_isa_table.table.columns[0]]
                )
            else:
                model.samples[study_item.file_name].table.row_count = 0

            model.parser_messages[study_item.file_name].extend(
                self.filter_messages(messages)
            )
            for assay_item in study_item.study_assays.assays:
                file_path = os.path.join(folder, assay_item.file_name)
                assay_isa_table_sheet, messages = parse_isa_table_sheet_from_fs(
                    file_path,
                    assay_file_expected_patterns,
                    offset=assay_sheet_offset,
                    limit=assay_sheet_limit,
                    fix_unicode_exceptions=True,
                )
                assay_isa_table: IsaTableFile = assay_isa_table_sheet
                model.parser_messages[assay_item.file_name].extend(
                    self.filter_messages(messages)
                )
                model.assays[assay_item.file_name].table = assay_isa_table.table
                model.assays[assay_item.file_name].sha256_hash = (
                    assay_isa_table.sha256_hash
                )
                model.assays[assay_item.file_name].file_path = assay_isa_table.file_path
                if assay_isa_table.table.data:
                    model.assays[assay_item.file_name].table.row_count = len(
                        assay_isa_table.table.data[assay_isa_table.table.columns[0]]
                    )
                    model.assays[assay_item.file_name].number_of_assay_rows = (
                        model.assays[assay_item.file_name].table.row_count
                    )
                else:
                    model.assays[assay_item.file_name].table.row_count = 0
        if not samples_sheet_limit or samples_sheet_limit > 0:
            model.has_sample_table_data = True
        if not assay_sheet_limit or assay_sheet_limit > 0:
            model.has_assay_table_data = True
        return model

    def get_phase3_input_data(
        self,
        study_id: str,
        folder: str,
        connection,
        assignment_sheet_offset: Union[int, None] = None,
        assignment_sheet_limit: Union[int, None] = None,
        model: MetabolightsStudyModel = None,
    ) -> MetabolightsStudyModel:
        if not model:
            model = self.get_phase1_input_data(study_id, folder, connection)

        for assignment_file in model.metabolite_assignments:
            absolute_path = os.path.join(folder, assignment_file)
            (
                metabolite_assignment_isa_table_sheet,
                messages,
            ) = parse_isa_table_sheet_from_fs(
                absolute_path,
                offset=assignment_sheet_offset,
                limit=assignment_sheet_limit,
                fix_unicode_exceptions=True,
            )
            metabolite_assignment_isa_table: IsaTableFile = (
                metabolite_assignment_isa_table_sheet
            )
            model.parser_messages[assignment_file].extend(
                self.filter_messages(messages)
            )
            metabolite_assignment_isa_table.file_path = assignment_file
            model.metabolite_assignments[assignment_file].table = (
                metabolite_assignment_isa_table.table
            )
            model.metabolite_assignments[assignment_file].sha256_hash = (
                metabolite_assignment_isa_table.sha256_hash
            )
            model.metabolite_assignments[assignment_file].file_path = assignment_file

            if metabolite_assignment_isa_table.table.data:
                model.metabolite_assignments[assignment_file].table.row_count = len(
                    metabolite_assignment_isa_table.table.data[
                        metabolite_assignment_isa_table.table.columns[0]
                    ]
                )
            else:
                model.metabolite_assignments[assignment_file].table.row_count = 0
        if not assignment_sheet_limit or assignment_sheet_limit > 0:
            model.has_assignment_table_data = True

        return model

    def get_phase4_input_data(
        self,
        study_id: str,
        folder: Union[str, None] = None,
        connection=None,
        model: MetabolightsStudyModel = None,
        calculate_data_folder_size: bool = False,
        calculate_metadata_size: bool = False,
    ) -> MetabolightsStudyModel:
        if not model:
            model = self.get_phase2_input_data(study_id, folder, connection)
        try:
            metadata, messages = self.folder_metadata_collector.get_folder_metadata(
                folder,
                calculate_data_folder_size=calculate_data_folder_size,
                calculate_metadata_size=calculate_metadata_size,
            )

            if messages:
                model.folder_reader_messages.extend(messages)
            if metadata:
                model.study_folder_metadata = metadata
                model.has_folder_metadata = True
        except Exception as ex:
            msg = f"Error while reading folder metadata: {str(ex)}"
            model.folder_reader_messages.append(
                GenericMessage(
                    type=GenericMessageType.ERROR,
                    short=msg,
                    detail=msg,
                )
            )
        return model

    def get_metabolights_study_model(
        self,
        study_id: str,
        folder: str,
        connection,
        load_sample_file: bool = True,
        load_assay_files: bool = True,
        load_maf_files: bool = True,
        load_folder_metadata: bool = True,
        samples_sheet_offset: Union[int, None] = None,
        samples_sheet_limit: Union[int, None] = None,
        assay_sheet_offset: Union[int, None] = None,
        assay_sheet_limit: Union[int, None] = None,
        assignment_sheet_offset: Union[int, None] = None,
        assignment_sheet_limit: Union[int, None] = None,
        calculate_data_folder_size: bool = False,
        calculate_metadata_size: bool = False,
    ) -> MetabolightsStudyModel:
        return self.load_study(
            study_id,
            folder,
            connection,
            load_sample_file=load_sample_file,
            load_assay_files=load_assay_files,
            load_maf_files=load_maf_files,
            load_folder_metadata=load_folder_metadata,
            samples_sheet_offset=samples_sheet_offset,
            samples_sheet_limit=samples_sheet_limit,
            assay_sheet_offset=assay_sheet_offset,
            assay_sheet_limit=assay_sheet_limit,
            assignment_sheet_offset=assignment_sheet_offset,
            assignment_sheet_limit=assignment_sheet_limit,
            calculate_data_folder_size=calculate_data_folder_size,
            calculate_metadata_size=calculate_metadata_size,
        )

    def load_study(
        self,
        study_id: str,
        study_path: str,
        connection=None,
        load_sample_file: bool = False,
        load_assay_files: bool = False,
        load_maf_files: bool = False,
        load_folder_metadata: bool = False,
        samples_sheet_offset: Union[int, None] = None,
        samples_sheet_limit: Union[int, None] = None,
        assay_sheet_offset: Union[int, None] = None,
        assay_sheet_limit: Union[int, None] = None,
        assignment_sheet_offset: Union[int, None] = None,
        assignment_sheet_limit: Union[int, None] = None,
        calculate_data_folder_size: bool = False,
        calculate_metadata_size: bool = False,
    ) -> MetabolightsStudyModel:

        model = self.get_phase1_input_data(study_id, study_path, connection)
        if load_sample_file and not load_assay_files and not load_folder_metadata:
            model = self.get_sample_file_input(
                study_id,
                study_path,
                connection,
                model=model,
                samples_sheet_offset=samples_sheet_offset,
                samples_sheet_limit=samples_sheet_limit,
            )

        if load_assay_files or load_folder_metadata:
            model = self.get_phase2_input_data(
                study_id,
                study_path,
                connection,
                model=model,
                samples_sheet_offset=samples_sheet_offset,
                samples_sheet_limit=samples_sheet_limit,
                assay_sheet_offset=assay_sheet_offset,
                assay_sheet_limit=assay_sheet_limit,
            )

        if load_folder_metadata:
            model = self.get_phase4_input_data(
                study_id,
                study_path,
                connection=connection,
                model=model,
                calculate_data_folder_size=calculate_data_folder_size,
                calculate_metadata_size=calculate_metadata_size,
            )
        else:
            message = GenericMessage(
                type=GenericMessageType.WARNING,
                short=f"Folder metadata collection is skipped for {study_id}",
                detail=f"Folder metadata collection is skipped for {study_id}",
            )
            model.folder_reader_messages.append(message)

        if load_maf_files:
            model = self.get_phase3_input_data(
                study_id,
                study_path,
                connection,
                model=model,
                assignment_sheet_limit=assignment_sheet_limit,
                assignment_sheet_offset=assignment_sheet_offset,
            )

        return model

    def update_investigation_file(
        self, model: MetabolightsStudyModel, folder, file_name="i_Investigation.txt"
    ):
        file = os.path.join(folder, file_name)
        investigation, messages = parse_investigation_from_fs(
            file, fix_unicode_exceptions=True
        )
        if not investigation:
            investigation = Investigation()

        model.investigation_file_path = file_name
        model.investigation = investigation
        model.parser_messages[file_name] = self.filter_messages(messages)
        model.has_investigation_data = True
        return model

    def update_study_db_metadata(
        self,
        model: MetabolightsStudyModel,
        collector: AbstractDbMetadataCollector,
        study_id,
        connection,
    ):
        try:
            study_db_metadata, messages = collector.get_study_metadata_from_db(
                study_id, connection
            )
            if messages:
                model.db_reader_messages.extend(messages)
            if study_db_metadata:
                model.study_db_metadata = study_db_metadata
                model.has_db_metadata = True
        except Exception as ex:
            message = GenericMessage(
                type=GenericMessageType.ERROR,
                short=f"Error while reading database for {study_id} metadata",
                detail=f"{str(ex)}",
            )
            model.db_reader_messages.append(message)
        return model

    def filter_messages(self, messages: List[ParserMessage]):
        return [
            f
            for f in messages
            if f.type == ParserMessageType.CRITICAL
            or f.type == ParserMessageType.ERROR
            or f.type == ParserMessageType.WARNING
        ]


if __name__ == "__main__":

    provider = MetabolightsStudyProvider()
    data = provider.load_study(
        "MTBLS1",
        study_path="tests/test-data/MTBLS1",
        load_assay_files=True,
        assay_sheet_limit=1,
    )
    print(data.parser_messages)



import json
import os
from typing import Union

import click

from metabolights_utils.commands.submission.model import (
    FtpLoginCredentials,
    LoginCredentials,
    RestApiCredentials,
)
from metabolights_utils.provider import definitions


@click.command(name="login")
@click.option(
    "--credentials_file_path",
    "-c",
    default=definitions.default_local_submission_credentials_file_path,
    help="Credentials file path.",
)
@click.option(
    "--rest_api_base_url",
    "-a",
    default=f"{definitions.default_rest_api_url}",
    help=f"MetaboLights Rest API base URL.",
)
@click.option(
    "--ftp_server_url",
    "-f",
    default=definitions.default_private_ftp_server_url,
    help=f"FTP server URL where MetaboLights repository is hosted. ",
)
@click.option("--ftp_user", prompt=True, default="", hide_input=False)
@click.option("--ftp_user_password", prompt=True, default="", hide_input=True)
@click.option("--api_token", prompt=True, default="", hide_input=True)
def submission_login(
    credentials_file_path: str = "",
    rest_api_base_url="",
    ftp_server_url: str = "",
    ftp_user: str = "",
    ftp_user_password: str = "",
    api_token: str = "",
):
    """
    Creates a file path to use connect private FTP server and MetaboLights Rest API.

    credentials_file_path (optional): File path to store user private FTP and MetaboLights Rest API credentials.

    """
    parent_directory = os.path.dirname(credentials_file_path)
    credentials: Union[None, LoginCredentials] = None
    initial: Union[None, LoginCredentials] = None
    try:
        if os.path.exists(credentials_file_path):
            with open(credentials_file_path, "r") as f:
                data = json.load(f)
                credentials = LoginCredentials.model_validate(data)
                initial = LoginCredentials.model_validate(data)
    except Exception as ex:
        click.echo("Error loading current credentials file. It will be overrided.")

    if not credentials:
        os.makedirs(parent_directory, exist_ok=True)
        credentials = LoginCredentials()
    try:
        if ftp_user and ftp_user_password:
            if ftp_server_url not in credentials.ftp_login:
                credentials.ftp_login[ftp_server_url] = FtpLoginCredentials()

            detail = credentials.ftp_login[ftp_server_url]
            detail.user_name = ftp_user
            detail.password = ftp_user_password

        if api_token:
            if rest_api_base_url not in credentials.rest_api_credentials:
                credentials.rest_api_credentials[rest_api_base_url] = (
                    RestApiCredentials()
                )
            detail = credentials.rest_api_credentials[rest_api_base_url]
            detail.api_token = api_token

        if api_token or ftp_user or ftp_user_password:
            json_data = credentials.model_dump()
            with open(credentials_file_path, "w") as f:
                json.dump(json_data, f, indent=4)
            click.echo("User credentials are updated succesfully.")
        else:
            click.echo("User credentials are not updated.")
    except Exception as ex:
        click.echo(f"Error while login: {str(ex)}")
        if initial:
            os.makedirs(parent_directory, exist_ok=True)
            json_data = initial.model_dump_json(indent=4)
            with open(credentials_file_path, "w") as f:
                json.dump(json_data, f, indent=4)


if __name__ == "__main__":
    submission_login()



from typing import Dict, List

from pydantic import Field
from typing_extensions import Annotated

from metabolights_utils.models.isa.common import AssayTechnique, IsaTableFile


class AssignmentFile(IsaTableFile):
    identified_metabolite_names: Annotated[
        List[str], Field(description="Names of metabolites defined in the ISA table.")
    ] = []
    metabolite_assignments: Annotated[
        Dict[str, str],
        Field(
            description="Metabolite names and assigned CHEBI id values in the ISA table."
        ),
    ] = {}
    assay_technique: Annotated[
        AssayTechnique,
        Field(
            description="Assay technique of assays that reference the current m_*.tsv file."
        ),
    ] = AssayTechnique()
    number_of_rows: Annotated[
        int, Field(description="Number of rows in ISA table.")
    ] = 0
    number_of_assigned_rows: Annotated[
        int, Field(description="Number of rows assigned a CHEBI id.")
    ] = 0
    number_of_unassigned_rows: Annotated[
        int, Field(description="Number of rows not assigned a CHEBI id.")
    ] = 0



import csv
import re
import sys
from functools import reduce
from io import IOBase
from typing import Dict, List, Union

from metabolights_utils.models.common import MetabolightsBaseModel
from metabolights_utils.models.isa.common import (
    INVESTIGATION_FILE_INITIAL_ROWS_SET,
    INVESTIGATION_FILE_STUDY_ROWS_SET,
)
from metabolights_utils.models.parser.common import ParserMessage
from metabolights_utils.models.parser.enums import ParserMessageType
from metabolights_utils.tsv.filter import Filter, FilterRegistry, TsvFileFilterOption
from metabolights_utils.tsv.sort import Sorter, SorterRegistry, TsvFileSortOption


def read_investigation_file(file_buffer: IOBase, messages: List[ParserMessage]):
    new_lines = read_investigation_file_lines(file_buffer, messages)
    dict_format = {}
    for i in range(len(new_lines)):
        detail = {}
        for j in range(len(new_lines[i])):
            detail[j] = new_lines[i][j]
        dict_format[i] = detail

    return dict_format


def read_investigation_file_lines(file_buffer: IOBase, messages: List[ParserMessage]):
    new_lines: List[List[str]] = []
    try:
        lines = file_buffer.readlines()

        content = "".join(lines)
        line_seperation = content.split("\n")
        richtext_lines = None
        tracked_lines = []
        for token in line_seperation:
            if not token:
                continue
            cleared_item = token.split("\t")[0]
            if (
                cleared_item in INVESTIGATION_FILE_STUDY_ROWS_SET
                or cleared_item in INVESTIGATION_FILE_INITIAL_ROWS_SET
                or cleared_item.startswith("Comment")
            ):
                if richtext_lines:
                    if not tracked_lines:
                        tracked_lines.append("")
                    tracked_lines[-1] = (
                        tracked_lines[-1] + "\n" + "\n".join(richtext_lines)
                    )
                    richtext_lines = []

                tracked_lines.append(token.replace("Comment ", "Comment", 1))
            else:
                if not richtext_lines:
                    richtext_lines = []
                richtext_lines.append(token)
        if richtext_lines:
            message = ParserMessage(type=ParserMessageType.WARNING)
            lines = {"\n".join(richtext_lines)}
            message.short = f"Invalid end of file. {lines}"
            message.detail = "Returned result is not complete."
            messages.append(message)

        for line in tracked_lines:
            result = line.split("\t")
            cleared_lines: List[str] = []

            first_item = result[0]
            for item in result:
                complete_line = False
                if cleared_lines:
                    previous: str = cleared_lines[-1]
                    if (
                        previous.startswith('"')
                        and len(previous) > 1
                        and previous.endswith('"')
                    ) or (not previous.startswith('"')):
                        complete_line = True
                else:
                    complete_line = True
                if complete_line:
                    cleared_lines.append(item)
                else:
                    cleared_lines[-1] = cleared_lines[-1] + "\t" + item

            result = cleared_lines
            result.remove(result[0])

            values = []
            values.append(first_item)
            for item in result:
                value = re.sub(r"\s+", " ", item).strip('"').strip()
                if value != item.strip('"'):
                    old = re.sub(r"\n", "<!new line char>", item)
                    old = re.sub(r"\t", "<!tab char>", old)

                    message = ParserMessage(type=ParserMessageType.WARNING)
                    short = "\n".join(
                        [
                            f"Line '{first_item}' is updated. Whitespace characters are removed",
                            f"Old: '{old}'",
                            f"New: '{value}'",
                        ]
                    )
                    message.short = short
                    message.detail = "Returned result is not complete."
                    messages.append(message)
                values.append(value.strip('"'))
            new_lines.append(values)
    except Exception as exc:
        message = ParserMessage(type=ParserMessageType.CRITICAL)
        message.short = f"The file can not read successfully. {str(exc)}"
        message.detail = f"Returned result is not complete. {str(exc)}"
        messages.append(message)

    for item in new_lines:
        new_line: List[str] = item
        if len(new_line) == 1 and not new_line[0].isupper():
            new_line.append('""')
    return new_lines


class TsvColumn(MetabolightsBaseModel):
    column_index: int = 0
    column_name: str = ""
    column_header: str = ""
    rows: Dict[int, str] = {}


class SelectedTsvFileContent(MetabolightsBaseModel):
    columns: List[TsvColumn] = []
    total_rows: int = 0
    total_filtered_rows: int = 0
    offset: int = 0
    limit: int = 0
    selected_column_count: int = 0
    total_columns: int = 0
    filter_options: List[TsvFileFilterOption] = []
    sort_options: List[TsvFileSortOption] = []


def read_table_file(
    file_buffer,
    messages: List[ParserMessage],
    selected_columns: Union[None, List[str]] = None,
    offset: Union[int, None] = None,
    limit: Union[int, None] = None,
    filter_options: List[TsvFileFilterOption] = None,
    sort_options: List[TsvFileSortOption] = None,
) -> SelectedTsvFileContent:
    file_buffer.seek(0)
    total_line = sum(1 for _ in file_buffer)
    total_data_rows = total_line - 1
    file_buffer.seek(0)
    reader = csv.reader(file_buffer, delimiter="\t")
    if total_line == 0:
        raise ValueError("There is no row in file")
    content: SelectedTsvFileContent = SelectedTsvFileContent()
    content.total_rows = total_data_rows
    content.filter_options = filter_options if filter_options else []
    content.sort_options = sort_options if sort_options else []
    if filter_options or sort_options:
        return read_table_file_with_filter_and_sort_option(
            reader,
            content,
            messages,
            selected_columns,
            offset,
            limit,
            filter_options,
            sort_options,
        )

    content.total_filtered_rows = total_data_rows
    offset = 0 if not offset else offset

    content.offset = (
        offset if offset < content.total_filtered_rows else content.total_filtered_rows
    )

    if content.offset < 0:
        content.offset = 0
    remaining_row_count = content.total_filtered_rows - content.offset

    if limit is None:
        content.limit = (
            remaining_row_count
            if remaining_row_count <= content.total_filtered_rows
            else content.total_filtered_rows
        )
    else:
        content.limit = remaining_row_count if remaining_row_count < limit else limit

    if content.limit < 0:
        content.limit = 0

    columns: Dict[str, TsvColumn] = {}
    column_indices: Dict[int, str] = {}
    selected_column_indices: Dict[int, str] = {}
    column_name_indices: Dict[str, int] = {}

    try:
        next_row_index = 0
        skipped_rows = 0
        read_rows = 0
        for row in reader:
            next_row_index += 1
            row_index = next_row_index - 1
            if row_index == 0:
                read_tsv_file_header(
                    content,
                    row,
                    selected_columns,
                    columns,
                    column_indices,
                    column_name_indices,
                    selected_column_indices,
                )
            else:
                if offset and skipped_rows < offset:
                    skipped_rows += 1
                    continue
                read_next_row = True
                if isinstance(limit, int) and read_rows >= limit:
                    read_next_row = False
                if read_next_row:
                    read_rows += 1
                    add_tsv_file_data_row(
                        row,
                        row_index - 1,
                        columns,
                        selected_column_indices=selected_column_indices,
                    )
                else:
                    break

    except Exception as exc:
        message = ParserMessage(type=ParserMessageType.CRITICAL)
        message.short = "ISA table file can not be read successfully."
        message.detail = f"Returned result is not complete. {str(exc)}"
        messages.append(message)
        return SelectedTsvFileContent()
    return content


def read_table_file_with_filter_and_sort_option(
    reader,
    content: SelectedTsvFileContent,
    messages: List[ParserMessage],
    selected_columns: Union[None, List[str]] = None,
    offset: Union[int, None] = None,
    limit: Union[int, None] = None,
    filter_options: List[TsvFileFilterOption] = None,
    sort_options: List[TsvFileSortOption] = None,
) -> SelectedTsvFileContent:
    columns: Dict[str, TsvColumn] = {}
    column_indices: Dict[int, str] = {}
    selected_column_indices: Dict[int, str] = {}
    column_name_indices: Dict[str, int] = {}
    filters: List[Filter] = []

    try:
        next_row_index = 0
        filtered_rows = []
        for row in reader:
            next_row_index += 1
            row_index = next_row_index - 1
            if row_index == 0:
                read_tsv_file_header(
                    content,
                    row,
                    selected_columns,
                    columns,
                    column_indices,
                    column_name_indices,
                    selected_column_indices,
                )
                if filter_options:
                    for filter_option in filter_options:
                        filter_option.search_columns = (
                            filter_option.search_columns
                            if filter_option.search_columns
                            else []
                        )
                        selected_filter: Filter = FilterRegistry.get_filter(
                            filter_option, column_name_indices, column_indices
                        )
                        filters.append(selected_filter)
                    # empty search column filters will be moved to end
                    filters.sort(
                        key=lambda x: (
                            len(x.filter_option.search_columns)
                            if x.filter_option.search_columns
                            else sys.maxsize
                        )
                    )
            else:
                if not filter_options:
                    filtered_rows.append((row_index - 1, row))
                else:
                    select: bool = True
                    for selected_filter in filters:
                        select = selected_filter.filter(row)
                        if not select:
                            break
                    if select:
                        filtered_rows.append((row_index - 1, row))
        reader = None

        sorters: List[Sorter] = []
        if sort_options:
            for sort_option in sort_options:
                col_index = column_name_indices[sort_option.column_name]
                sorter: Sorter = SorterRegistry.get_sorter(
                    sort_option,
                    col_index,
                    column_name_indices,
                    column_indices,
                )
                sorters.append(sorter)
        if sorters:
            filtered_rows = reduce(
                lambda s, sorter: sorted(
                    s, key=sorter.sort, reverse=sorter.sort_option.reverse
                ),
                reversed(sorters),
                filtered_rows,
            )

        content.total_filtered_rows = len(filtered_rows)
        offset = 0 if not offset else offset

        content.offset = (
            offset
            if offset < content.total_filtered_rows
            else content.total_filtered_rows
        )
        remaining_row_count = content.total_filtered_rows - content.offset
        if limit is None:
            content.limit = (
                remaining_row_count
                if remaining_row_count <= content.total_filtered_rows
                else content.total_filtered_rows
            )
        else:
            content.limit = (
                remaining_row_count if remaining_row_count < limit else limit
            )

        skipped_rows = 0
        read_rows = 0
        for data_row_index, row in filtered_rows:
            if offset and skipped_rows < offset:
                skipped_rows += 1
                continue
            read_next_row = True
            if isinstance(limit, int) and read_rows >= limit:
                read_next_row = False
            if read_next_row:
                read_rows += 1
                add_tsv_file_data_row(
                    row,
                    data_row_index,
                    columns,
                    selected_column_indices=selected_column_indices,
                )
            else:
                break

    except Exception as exc:
        message = ParserMessage(type=ParserMessageType.CRITICAL)
        message.short = "ISA table file can not be read successfully."
        message.detail = f"Returned result is not complete. {str(exc)}"
        messages.append(message)
    return content


def add_tsv_file_data_row(
    data_row,
    row_index: int,
    columns: Dict[str, TsvColumn],
    selected_column_indices: Dict[int, str],
):
    for column_index in selected_column_indices:
        column = columns[selected_column_indices[column_index]]
        column.rows[row_index] = data_row[column_index]


def prepare_column_names(
    selected_column_names: Union[None, List[str]],
    header_row: List[str],
    column_indices: Dict[int, str],
    column_name_indices: Dict[str, int],
):
    column_headers: Dict[str, int] = {}
    col_index = 0
    for col_index in range(len(header_row)):
        column_header = header_row[col_index]
        if column_header not in column_headers:
            column_headers[column_header] = 1
        else:
            column_headers[column_header] = column_headers[column_header] + 1

        column_name = (
            f"{column_header}.{(column_headers[column_header] - 1)}"
            if column_headers[column_header] > 1
            else column_header
        )
        column_name_indices[column_name] = col_index
        column_indices[col_index] = column_name
    new_ordered_columns = []
    invalid_column_names = []
    if selected_column_names is not None:
        for name in selected_column_names:
            if name in new_ordered_columns:
                continue
            if name == "Unit" or name.startswith("Unit."):
                continue
            elif name == "Term Source REF" or name.startswith("Term Source REF."):
                continue
            elif name == "Term Accession Number" or name.startswith(
                "Term Accession Number."
            ):
                continue
            elif name not in column_name_indices:
                invalid_column_names.append(name)
                continue
            else:
                new_ordered_columns.append(name)
                col_index = column_name_indices[name]
                term_source_relative_index = 1
                if col_index + 1 < len(header_row):
                    next_item: str = column_indices[col_index + 1]
                    if next_item == "Unit" or next_item.startswith("Unit."):
                        new_ordered_columns.append(next_item)
                        term_source_relative_index = 2

                if col_index + term_source_relative_index + 1 < len(header_row):
                    next_item: str = column_indices[
                        col_index + term_source_relative_index
                    ]
                    if next_item == "Term Source REF" or next_item.startswith(
                        "Term Source REF."
                    ):
                        new_ordered_columns.append(next_item)
                        next_item: str = column_indices[
                            col_index + term_source_relative_index + 1
                        ]
                        if (
                            next_item == "Term Accession Number"
                            or next_item.startswith("Term Accession Number.")
                        ):
                            new_ordered_columns.append(next_item)
        selected_column_names.clear()
        selected_column_names.extend(new_ordered_columns)
    if invalid_column_names:
        raise TypeError(
            "Column(s) do(es) not exist: " + ", ".join(invalid_column_names)
        )


def read_tsv_file_header(
    content: SelectedTsvFileContent,
    header_row,
    column_names: Union[None, List[str]],
    columns: Dict[str, TsvColumn],
    column_indices: Dict[int, str],
    column_name_indices: Dict[str, int],
    selected_column_indices: Dict[str, int],
):
    prepare_column_names(column_names, header_row, column_indices, column_name_indices)
    content.selected_column_count = len(column_names) if column_names else 0
    content.total_columns = len(header_row) if header_row else 0

    try:
        if column_names:
            for column_name in column_names:
                col_index = column_name_indices[column_name]
                column_header = header_row[col_index]
                column = TsvColumn(
                    column_index=col_index,
                    column_header=column_header,
                    column_name=column_name,
                )
                columns[column_name] = column
                selected_column_indices[col_index] = column_name
                content.columns.append(column)
        else:
            for col_index in range(len(header_row)):
                column_name = column_indices[col_index]
                column_header = header_row[col_index]
                column = TsvColumn(
                    column_index=col_index,
                    column_header=column_header,
                    column_name=column_name,
                )
                columns[column_name] = column
                selected_column_indices[col_index] = column_name
                content.columns.append(column)

        # if column_names:
        #     invalid_column_names = []
        #     for selected_column in column_names:
        #         if selected_column not in columns:
        #             invalid_column_names.append(selected_column)
        #     if invalid_column_names:
        #         raise TypeError(f"Column(s) do(es) not exist: " + ", ".join(invalid_column_names))
    except Exception as exc:
        raise exc



import pathlib
import uuid
from typing import Dict, List

from metabolights_utils.tsv import model as actions
from metabolights_utils.tsv.actions.base import BaseTsvAction


class DeleteColumnsTsvAction(BaseTsvAction):
    def apply_action(
        self,
        source_file_path: pathlib.Path,
        target_file_path: pathlib.Path,
        action: actions.TsvDeleteColumnsAction,
        read_encoding: str = "utf-8",
        write_encoding: str = "utf-8",
    ) -> actions.TsvActionResult:
        result: actions.TsvActionResult = actions.TsvActionResult(action=action)
        if action.action_type != actions.TsvActionType.DELETE_COLUMN:
            result.message = "Action name is not valid"
            return result

        action: actions.TsvDeleteColumnsAction = action

        columns: Dict[int, str] = (
            action.current_columns if action.current_columns else {}
        )
        if not columns:
            result.message = "There is not column index"
            return result

        column_indices: List[int] = list(columns.keys()).copy()
        column_indices.sort()

        if not action.id:
            uuid_value = str(uuid.uuid4().hex)
            action.id = uuid_value

        try:
            with open(source_file_path, "r", encoding=read_encoding) as source:
                header_line = source.readline()
                header_names = header_line.strip("\n").split("\t")
                new_header_names = []
                for column_idx, value in enumerate(header_names):
                    if column_idx in column_indices:
                        header_name = columns[column_idx]
                        if header_name != value:
                            result.message = f"Input header name does not match the actual one for index {column_idx}. Expected: {header_name}, found: {value}"
                            return result
                        continue

                    new_header_names.append(value)

                with open(target_file_path, "w", encoding=write_encoding) as target:
                    self.write_row(target, new_header_names)
                    for line in source:
                        row = line.strip("\n").split("\t")
                        new_row = [
                            x[1] for x in enumerate(row) if x[0] not in column_indices
                        ]
                        new_row.append(value)
                        self.write_row(target, new_row)
            result.success = True
        except Exception as exc:
            result.message = f"{str(exc)}"

        return result



import sys

import click

from metabolights_utils.commands.model.model_cli import model_cli
from metabolights_utils.commands.public.public_cli import public_cli
from metabolights_utils.commands.submission.submission_cli import submission_cli


@click.group()
def cli():
    pass


cli.add_command(public_cli)
cli.add_command(submission_cli)
cli.add_command(model_cli)

if __name__ == "__main__":
    if len(sys.argv) == 0:
        cli(["--help"])
    else:
        cli()



import re
from typing import Dict, List, Set

from pydantic import BaseModel, Field
from typing_extensions import Annotated

ftp_list = r"([\-\w]+)\s+(\d+)\s+(\w+)\s+(\w+)\s+(\w+)\s+(\w+\s+\w+\s+[\w:]+)\s+(.*)\s*"
ftp_list_response_pattern = re.compile(ftp_list)


class FtpResponse(BaseModel):
    success: bool = False
    code: int = -1
    message: str = ""


class FtpFiles(FtpResponse):
    files: List[str] = []


class LocalDirectory(FtpResponse):
    root_path: str = ""
    local_files: List[str] = []
    local_folders: List[str] = []
    actions: Dict[str, str] = {}


class FtpFileDescriptor(BaseModel):
    base_name: Annotated[str, Field(description="")] = ""
    size_in_bytes: Annotated[int, Field(description="")] = 0
    is_directory: Annotated[bool, Field(description="")] = False
    is_link: Annotated[bool, Field(description="")] = False
    mode: Annotated[str, Field(description="")] = ""


class FtpFolderContent(FtpFiles):
    source_directory: str = ""
    folders: List[str] = []
    links: Set[str] = set()
    descriptors: List[FtpFileDescriptor] = []

    def parse_line(self, line: str) -> None:
        if line and line.strip():
            result = ftp_list_response_pattern.search(line)
            if result and result.groups():
                groups = result.groups()
                mode = groups[0][1:]
                binary_mode = re.sub(r"[^0]", "1", mode.replace("-", "0"))
                octal_mode = oct(int(binary_mode, 2)).replace("0o", "")
                filename = groups[6]
                size = groups[4]
                size_in_bytes = 0
                if size.isnumeric():
                    size_in_bytes = int(size)
                descriptor = FtpFileDescriptor(
                    mode=octal_mode, base_name=filename, size_in_bytes=size_in_bytes
                )

                if groups[0].startswith("l"):
                    descriptor.is_link = True
                    self.links.add(filename)
                elif groups[0].startswith("d"):
                    self.folders.append(groups[6])
                    descriptor.is_directory = True
                else:
                    self.files.append(groups[6])
                self.descriptors.append(descriptor)



import re
from typing import Union

import unidecode


class MetabolightsFileNameUtils(object):
    @staticmethod
    def sanitise_filename(
        filename: str,
        replacement_chars_pattern: Union[None, str] = None,
        allow_spaces: bool = False,
    ) -> str:
        if not filename or not filename.strip():
            return ""

        if not replacement_chars_pattern:
            replacement_chars_pattern = "[^/a-zA-Z0-9_.-]"
        filename = unidecode.unidecode(filename.strip())
        filename = filename.replace("+", "_PLUS_")
        if not allow_spaces:
            filename = "_".join(
                [x.strip() for x in filename.strip().split() if x.strip()]
            )

        filename = re.sub(replacement_chars_pattern, "_", filename)
        return filename



import json
import os
from typing import Union

from metabolights_utils.commands.submission.model import (
    FtpLoginCredentials,
    LoginCredentials,
    RestApiCredentials,
)


def get_submission_private_ftp_credentials(
    credentials_file_path: str, private_ftp_url: str
) -> Union[None, FtpLoginCredentials]:
    credentials = get_submission_credentials(credentials_file_path)
    if credentials:
        if private_ftp_url in credentials.ftp_login:
            return credentials.ftp_login[private_ftp_url]
    return None


def get_submission_rest_api_credentials(
    credentials_file_path: str, rest_api_base_url: str
) -> Union[None, RestApiCredentials]:
    credentials = get_submission_credentials(credentials_file_path)
    if credentials:
        if rest_api_base_url in credentials.rest_api_credentials:
            return credentials.rest_api_credentials[rest_api_base_url]
    return None


def get_submission_credentials(credentials_file_path: str):
    credentials: Union[None, LoginCredentials] = None
    try:
        if os.path.exists(credentials_file_path):
            with open(credentials_file_path, "r") as f:
                data = json.load(f)
                credentials = LoginCredentials.model_validate(data)
    except Exception as ex:
        print(f"Error while loading {credentials_file_path}. {str(ex)}")
    return credentials



from typing import Dict, List, Literal, Union

from pydantic import BaseModel, ConfigDict, Field
from pydantic.alias_generators import to_snake
from typing_extensions import Annotated

from metabolights_utils.models.common import MetabolightsBaseModel
from metabolights_utils.models.isa.enums import ColumnsStructure
from metabolights_utils.tsv.filter import TsvFileFilterOption
from metabolights_utils.tsv.sort import TsvFileSortOption

INVESTIGATION_FILE_SECTION_NAMES = {
    "ONTOLOGY SOURCE REFERENCE",
    "INVESTIGATION",
    "INVESTIGATION PUBLICATIONS",
    "INVESTIGATION CONTACTS",
    "STUDY",
    "STUDY DESIGN DESCRIPTORS",
    "STUDY PUBLICATIONS",
    "STUDY FACTORS",
    "STUDY ASSAYS",
    "STUDY PROTOCOLS",
    "STUDY CONTACTS",
}

INVESTIGATION_FILE_INITIAL_ROWS = [
    "ONTOLOGY SOURCE REFERENCE",
    "Term Source Name",
    "Term Source File",
    "Term Source Version",
    "Term Source Description",
    "INVESTIGATION",
    "Investigation Identifier",
    "Investigation Title",
    "Investigation Description",
    "Investigation Submission Date",
    "Investigation Public Release Date",
    "INVESTIGATION PUBLICATIONS",
    "Investigation PubMed ID",
    "Investigation Publication DOI",
    "Investigation Publication Author List",
    "Investigation Publication Title",
    "Investigation Publication Status",
    "Investigation Publication Status Term Accession Number",
    "Investigation Publication Status Term Source REF",
    "INVESTIGATION CONTACTS",
    "Investigation Person Last Name",
    "Investigation Person First Name",
    "Investigation Person Mid Initials",
    "Investigation Person Email",
    "Investigation Person Phone",
    "Investigation Person Fax",
    "Investigation Person Address",
    "Investigation Person Affiliation",
    "Investigation Person Roles",
    "Investigation Person Roles Term Accession Number",
    "Investigation Person Roles Term Source REF",
]

INVESTIGATION_FILE_STUDY_ROWS = [
    "STUDY",
    "Study Identifier",
    "Study Title",
    "Study Description",
    "Study Submission Date",
    "Study Public Release Date",
    "Study File Name",
    "STUDY DESIGN DESCRIPTORS",
    "Study Design Type",
    "Study Design Type Term Accession Number",
    "Study Design Type Term Source REF",
    "STUDY PUBLICATIONS",
    "Study PubMed ID",
    "Study Publication DOI",
    "Study Publication Author List",
    "Study Publication Title",
    "Study Publication Status",
    "Study Publication Status Term Accession Number",
    "Study Publication Status Term Source REF",
    "STUDY FACTORS",
    "Study Factor Name",
    "Study Factor Type",
    "Study Factor Type Term Accession Number",
    "Study Factor Type Term Source REF",
    "STUDY ASSAYS",
    "Study Assay File Name",
    "Study Assay Measurement Type",
    "Study Assay Measurement Type Term Accession Number",
    "Study Assay Measurement Type Term Source REF",
    "Study Assay Technology Type",
    "Study Assay Technology Type Term Accession Number",
    "Study Assay Technology Type Term Source REF",
    "Study Assay Technology Platform",
    "STUDY PROTOCOLS",
    "Study Protocol Name",
    "Study Protocol Type",
    "Study Protocol Type Term Accession Number",
    "Study Protocol Type Term Source REF",
    "Study Protocol Description",
    "Study Protocol URI",
    "Study Protocol Version",
    "Study Protocol Parameters Name",
    "Study Protocol Parameters Name Term Accession Number",
    "Study Protocol Parameters Name Term Source REF",
    "Study Protocol Components Name",
    "Study Protocol Components Type",
    "Study Protocol Components Type Term Accession Number",
    "Study Protocol Components Type Term Source REF",
    "STUDY CONTACTS",
    "Study Person Last Name",
    "Study Person First Name",
    "Study Person Mid Initials",
    "Study Person Email",
    "Study Person Phone",
    "Study Person Fax",
    "Study Person Address",
    "Study Person Affiliation",
    "Study Person Roles",
    "Study Person Roles Term Accession Number",
    "Study Person Roles Term Source REF",
]

INVESTIGATION_FILE_INITIAL_ROWS_SET = set(INVESTIGATION_FILE_INITIAL_ROWS)
INVESTIGATION_FILE_STUDY_ROWS_SET = set(INVESTIGATION_FILE_STUDY_ROWS)


class IsaTabConfig(MetabolightsBaseModel):
    field_order: Annotated[Union[None, List[str]], Field(description="")] = None
    section_header: Annotated[Union[None, str], Field(description="")] = None
    section_prefix: Annotated[Union[None, str], Field(description="")] = None
    specification_version: Annotated[str, Field(description="")] = ""
    specification_date: Annotated[str, Field(description="")] = ""


class IsaAbstractModel(MetabolightsBaseModel):
    isatab_config: Annotated[
        IsaTabConfig,
        Field(exclude=True),
    ] = IsaTabConfig()

    model_config = ConfigDict(extra="forbid")

    @classmethod
    def get_attribute(cls, model: BaseModel, attribute_name):
        model_attribute_key = to_snake(attribute_name)
        return getattr(model, model_attribute_key)

    @classmethod
    def is_empty_model(cls, model: BaseModel):
        schema = model.model_json_schema()
        for item_key in schema["properties"]:
            item_field_definition = schema["properties"][item_key]
            if (
                not item_field_definition
                or "auto_fill" not in item_field_definition
                or not item_field_definition["auto_fill"]
            ):
                continue
            model_attribute_key = to_snake(item_key)
            item_val = cls.get_attribute(model, model_attribute_key)
            if isinstance(item_val, IsaAbstractModel):
                is_empty = IsaAbstractModel.is_empty_model(item_val)
                if not is_empty:
                    return False
            elif item_val:
                return False
        return True


class AssayTechnique(IsaAbstractModel):
    name: Annotated[
        str, Field(description="Technique name used to identify metabolites.")
    ] = ""
    main_technique: Annotated[
        Literal["MS", "NMR"],
        Field(description="Main category of metabolite identification technique."),
    ] = ""
    technique: Annotated[str, Field(description="Technique")] = ""
    sub_technique: Annotated[str, Field(description="Sub-techniique")] = ""

    def __str__(self) -> str:
        return self.name

    def __eq__(self, other):
        return self.__str__() == other.__str__()

    def __hash__(self):
        return hash(self.__str__())

    def __repr__(self):
        return self.__str__()


class OntologyItem(IsaAbstractModel):
    term: Annotated[str, Field(description="Ontology term.")] = ""
    term_source_ref: Annotated[
        str,
        Field(
            description="Source reference name of ontology term. e.g., EFO, OBO, NCIT."
        ),
    ] = ""
    term_accession_number: Annotated[
        str, Field(description="Accession number of ontology term.")
    ] = ""

    def __str__(self) -> str:
        if self.term:
            if self.term_source_ref or self.term_accession_number:
                return (
                    f"{self.term} [{self.term_source_ref}:{self.term_accession_number}]"
                )
            else:
                return self.term
        return ""

    def __eq__(self, other):
        return self.__str__() == other.__str__()

    def __hash__(self):
        return hash(self.__str__())

    def __repr__(self):
        return self.__str__()


class OrganismAndOrganismPartPair(IsaAbstractModel):
    organism: Annotated[OntologyItem, Field(description="Organism ontology value.")] = (
        OntologyItem()
    )
    organism_part: Annotated[
        OntologyItem, Field(description="Organism part ontology value.")
    ] = OntologyItem()
    variant: Annotated[OntologyItem, Field(description="Variant ontology value.")] = (
        OntologyItem()
    )
    sample_type: Annotated[
        OntologyItem, Field(description="Sample type ontology value.")
    ] = OntologyItem()

    def __str__(self) -> str:
        if self.organism and self.organism_part and self.variant and self.sample_type:
            return ":".join(
                [
                    self.organism.__str__(),
                    self.organism_part.__str__(),
                    self.variant.__str__(),
                    self.sample_type.__str__(),
                ]
            )

        if self.organism and self.organism_part and self.variant:
            return ":".join(
                [
                    self.organism.__str__(),
                    self.organism_part.__str__(),
                    self.variant.__str__(),
                ]
            )
        if self.organism and self.organism_part:
            return ":".join([self.organism.__str__(), self.organism_part.__str__()])
        if self.organism:
            return self.organism.__str__()

        return ""

    def __eq__(self, other):
        return self.__str__() == other.__str__()

    def __hash__(self):
        return hash(self.__str__())

    def __repr__(self):
        return self.__str__()


class NumericItem(OntologyItem):
    value: Annotated[str, Field(description="")] = ""
    unit: Annotated[OntologyItem, Field(description="")] = OntologyItem()


class ProtocolField(MetabolightsBaseModel):
    header_name: Annotated[str, Field(description="")] = ""


class ProtocolOntologyItem(ProtocolField):
    data: Annotated[List[OntologyItem], Field(description="")] = []


class ProtocolTextItem(ProtocolField):
    data: Annotated[List[str], Field(description="")] = []


class ProtocolNumericOntologyItem(ProtocolField):
    data: Annotated[List[NumericItem], Field(description="")] = []


class ProtocolItem(MetabolightsBaseModel):
    name: Annotated[str, Field(description="")] = ""
    text_fields: Annotated[Dict[str, ProtocolTextItem], Field(description="")] = {}
    numeric_fields: Annotated[
        Dict[str, ProtocolNumericOntologyItem], Field(description="")
    ] = {}
    ontology_fields: Annotated[
        Dict[str, ProtocolOntologyItem], Field(description="")
    ] = {}


class ProtocolFields(MetabolightsBaseModel):
    name: Annotated[str, Field(description="")] = ""
    text_fields: Annotated[Dict[str, ProtocolTextItem], Field(description="")] = {}
    numeric_fields: Annotated[
        Dict[str, ProtocolNumericOntologyItem], Field(description="")
    ] = {}
    ontology_fields: Annotated[
        Dict[str, ProtocolOntologyItem], Field(description="")
    ] = {}

    additional_text_fields: Annotated[
        List[ProtocolTextItem], Field(description="Additional text")
    ] = []
    additional_numeric_fields: Annotated[
        List[ProtocolNumericOntologyItem], Field(description="")
    ] = []
    additional_ontology_fields: Annotated[
        List[ProtocolOntologyItem], Field(description="")
    ] = []


class Comment(IsaAbstractModel):
    name: Annotated[str, Field(description="Name of comment")] = ""
    value: Annotated[List[str], Field(description="Values of comment")] = []


class IsaTableColumn(IsaAbstractModel):
    column_index: Annotated[
        Union[int, None],
        Field(description="Index of column in ISA Table. First column index is 0."),
    ] = None

    column_name: Annotated[
        str,
        Field(
            description="Unique name of column. It is same if column header is unique in ISA table."
        ),
    ] = ""

    column_header: Annotated[str, Field(description="Header of ISA table column.")] = ""

    additional_columns: Annotated[
        List[str],
        Field(
            description="Linked column names. If column is ontology or a column with unit ontology column, it lists the following column headers."
        ),
    ] = []

    column_category: Annotated[
        str,
        Field(
            description="Column category. e.g., Parameter Value, Factor Value, protocol"
        ),
    ] = ""

    column_structure: Annotated[
        ColumnsStructure, Field(description="Structure of column in ISA table.")
    ] = ColumnsStructure.SINGLE_COLUMN

    column_prefix: Annotated[
        str, Field(description="Column prefix if header has a value between [].")
    ] = ""

    column_search_pattern: Annotated[
        str,
        Field(
            description="Search regex expression to fetch value if if header has a value between []."
        ),
    ] = ""

    def __hash__(self):
        return hash(self.column_name)


class IsaTable(IsaAbstractModel):
    columns: Annotated[
        List[str],
        Field(
            description="Unique column names of ISA table. If there are dublicate column headers, column names are created with the following pattern: header name + '.X' where X is sequencial number for each duplicate header name."
        ),
    ] = []

    # column_baseline: Annotated[
    #     List[str],
    #     Field(description="Baseline (before modification) column names of this table."),
    # ] = []

    headers: Annotated[
        List[IsaTableColumn],
        Field(
            description="Metadata of ISA table columns. e.i., header name, column name, column index"
        ),
    ] = []

    data: Annotated[
        Dict[str, List[str]], Field(description="Data columns and their row values.")
    ] = {}

    # row_baseline: Annotated[
    #     List[int],
    #     Field(description="Baseline (before modification) rows of this table."),
    # ] = []

    row_indices: Annotated[
        List[int],
        Field(
            description="Current rows' positions in ISA table. e.g., [2, 3, 5]: First three rows of 'data' are 2nd, 3th and 5th rows in ISA Table."
        ),
    ] = []

    column_indices: Annotated[
        List[int], Field(description="Current columns' positions in ISA table.")
    ] = []

    filtered_total_row_count: Annotated[
        int, Field(description="Total row count after filter operations.")
    ] = 0

    row_offset: Annotated[
        int, Field(description="Skipped rows after filter operations.")
    ] = 0

    row_count: Annotated[
        int,
        Field(description="Row count listed in this table after filter operations."),
    ] = 0

    total_row_count: Annotated[
        int, Field(description="Total row count in ISA table file.")
    ] = 0

    selected_column_count: Annotated[
        int, Field(description="Number of selected columns in this table file.")
    ] = 0

    total_column_count: Annotated[
        int, Field(description="Number of all columns in ISA table file.")
    ] = 0

    filter_options: Annotated[
        List[TsvFileFilterOption],
        Field(description="Applied filters on ISA table file."),
    ] = []

    sort_options: Annotated[
        List[TsvFileSortOption],
        Field(description="Applied sort operations on ISA table file."),
    ] = []


class IsaTableFile(IsaAbstractModel):
    file_path: Annotated[
        str,
        Field(
            description="Relative file path of ISA table file (s_*.txt, a_*.txt, m_*.tsv)."
        ),
    ] = ""

    sha256_hash: Annotated[
        str,
        Field(description="SHA256 hash value of ISA table file."),
    ] = "e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855"

    table: Annotated[
        IsaTable,
        Field(description="All or partial content of ISA table file."),
    ] = IsaTable()



from enum import Enum
from typing import Dict, List, Union

from pydantic import ConfigDict, Field
from typing_extensions import Annotated

from metabolights_utils.common import CamelCaseModel
from metabolights_utils.models.common import GenericMessage
from metabolights_utils.models.isa.common import IsaAbstractModel
from metabolights_utils.models.metabolights.metabolights_study import (
    BaseMetabolightsStudyModel,
)


class UserStatus(str, Enum):
    NEW = "NEW"
    VERIFIED = "VERIFIED"
    ACTIVE = "ACTIVE"
    FROZEN = "FROZEN"

    @staticmethod
    def get_from_int(int_value: int):
        if int_value == 0:
            return UserStatus.NEW
        elif int_value == 1:
            return UserStatus.VERIFIED
        elif int_value == 2:
            return UserStatus.ACTIVE
        else:
            return UserStatus.FROZEN


class UserRole(str, Enum):
    SUBMITTER = "SUBMITTER"
    CURATOR = "CURATOR"
    ANONYMOUS = "ANONYMOUS"

    @staticmethod
    def get_from_int(int_value: int):
        if int_value == 0:
            return UserRole.SUBMITTER
        elif int_value == 1:
            return UserRole.CURATOR
        else:
            return UserRole.ANONYMOUS


class CurationRequest(str, Enum):
    MANUAL_CURATION = "MANUAL_CURATION"
    NO_CURATION = "NO_CURATION"
    SEMI_AUTOMATED_CURATION = "SEMI_AUTOMATED_CURATION"

    @staticmethod
    def get_from_int(int_value: int):
        if int_value == 0:
            return CurationRequest.MANUAL_CURATION
        elif int_value == 1:
            return CurationRequest.NO_CURATION
        elif int_value == 2:
            return CurationRequest.SEMI_AUTOMATED_CURATION
        return CurationRequest.MANUAL_CURATION


class StudyStatus(str, Enum):
    SUBMITTED = "SUBMITTED"
    INCURATION = "INCURATION"
    INREVIEW = "INREVIEW"
    PUBLIC = "PUBLIC"
    DORMANT = "DORMANT"

    @staticmethod
    def get_from_int(int_value: int):
        if int_value == 0:
            return StudyStatus.SUBMITTED
        elif int_value == 1:
            return StudyStatus.INCURATION
        elif int_value == 2:
            return StudyStatus.INREVIEW
        elif int_value == 3:
            return StudyStatus.PUBLIC
        else:
            return StudyStatus.DORMANT


class StudyFileDescriptor(IsaAbstractModel):
    file_path: Annotated[
        str, Field(description="File relative path. e.g., FILES/RAW_FILES/raw_data.d")
    ] = ""
    base_name: Annotated[
        str,
        Field(
            description="File basename. e.g., raw_data.d for FILES/RAW_FILES/raw_data.d"
        ),
    ] = ""
    parent_directory: Annotated[
        str,
        Field(
            description="Relative parent directory. e.g., FILES/RAW_FILES for FILES/RAW_FILES/raw_data.d"
        ),
    ] = ""
    extension: Annotated[
        str,
        Field(
            description="File extension. e.g., .d for raw_data.d for FILES/RAW_FILES/raw_data.d"
        ),
    ] = ""
    size_in_bytes: Annotated[int, Field(description="Size of file in bytes.")] = 0
    is_directory: Annotated[
        bool, Field(description="True if the file is a folder.")
    ] = False
    is_link: Annotated[
        bool, Field(description="True if the file is a symbolic link.")
    ] = False
    modified_at: Annotated[
        int,
        Field(
            description="Modified time of the file. The value is a timestamp in UTC."
        ),
    ] = 0
    created_at: Annotated[
        int,
        Field(
            description="Creation time of the file. The value is a timestamp in UTC."
        ),
    ] = 0
    mode: Annotated[
        str, Field(description="File permission mode in octal. e.g., 775")
    ] = ""
    tags: Annotated[List[str], Field(description="Tag values for the file.")] = []


class StudyFolderMetadata(IsaAbstractModel):
    data_folder_size_calculated: Annotated[
        bool, Field(description="True if data folder size is calculated.")
    ] = False
    metadata_folder_size_calculated: Annotated[
        bool, Field(description="True if metadata folder size is calculated.")
    ] = False
    folder_size_in_bytes: Annotated[
        int,
        Field(
            description="Study folder size in bytes. Folder size may be size of metadata files, data files or both depends on data_folder_size_calculated or metadata_folder_size_calculated."
        ),
    ] = -1
    folder_size_in_str: Annotated[
        str, Field(description="Human readable study folder size in MB or GB")
    ] = ""
    folders: Annotated[
        Dict[str, StudyFileDescriptor],
        Field(
            description="Folders and summary information about each one within a study folder."
        ),
    ] = {}
    files: Annotated[
        Dict[str, StudyFileDescriptor],
        Field(
            description="Files and summary information about each one within a study folder."
        ),
    ] = {}


class Submitter(CamelCaseModel):
    db_id: Annotated[int, Field(description="Database unique id for submitter.")] = -1
    orcid: Annotated[str, Field(description="ORCID of submitter.")] = ""
    address: Annotated[str, Field(description="Address of submitter.")] = ""
    join_date: Annotated[str, Field(description="Registration date of submitter.")] = "R"
    user_name: Annotated[
        str, Field(description="username (email address) of submitter.")
    ] = ""
    first_name: Annotated[str, Field(description="First name of submitter.")] = ""
    last_name: Annotated[str, Field(description="Last name of submitter.")] = ""
    affiliation: Annotated[str, Field(description="Submitter's affiliation.")] = ""
    affiliation_url: Annotated[
        str, Field(description="Submitter's affiliation webpage URL.")
    ] = ""
    status: Annotated[UserStatus, Field(description="Status of user.")] = (
        UserStatus.FROZEN
    )
    role: Annotated[UserRole, Field(description="Role of submmiter.")] = (
        UserRole.ANONYMOUS
    )

    model_config = ConfigDict(from_attributes=True)


class StudyDBMetadata(CamelCaseModel):
    db_id: Annotated[int, Field(description="Database unique id for study.")] = -1
    study_id: Annotated[str, Field(description="Study accession number.")] = ""
    numeric_study_id: Annotated[
        int,
        Field(
            description="Numarical part of study accession number. e.g., 123 for MTBLS123"
        ),
    ] = -1
    obfuscation_code: Annotated[
        str, Field(description="Obfusication code of study.")
    ] = ""
    study_size: Annotated[
        Union[None, int], Field(description="Total size (bytes) of study folder.")
    ] = -1
    submission_date: Annotated[str, Field(description="Study submission date.")] = ""
    release_date: Annotated[
        str, Field(description="Study public release date (actual or anticipated.)")
    ] = ""
    update_date: Annotated[str, Field(description="Last update date of study.")] = ""
    status_date: Annotated[
        str, Field(description="Last status update date of study.")
    ] = ""
    study_types: Annotated[
        List[str], Field(description="Study types defined by curators.")
    ] = []
    status: Annotated[StudyStatus, Field(description="Status of study.")] = (
        StudyStatus.DORMANT
    )
    curation_request: Annotated[
        CurationRequest,
        Field(
            description="Curation request. e.g., no curation request, manual curation request."
        ),
    ] = CurationRequest.MANUAL_CURATION
    overrides: Annotated[
        Dict[str, str], Field(description="Overrided validation rules.")
    ] = {}
    submitters: Annotated[
        List[Submitter], Field(description="Submitters of study.")
    ] = []


class MetabolightsStudyModel(BaseMetabolightsStudyModel):
    study_db_metadata: Annotated[
        StudyDBMetadata,
        Field(
            description="Basic metatada information (status, database id, etc.) stored in database."
        ),
    ] = StudyDBMetadata()

    study_folder_metadata: Annotated[
        StudyFolderMetadata,
        Field(description="Descriptors of files and folders within study folder."),
    ] = StudyFolderMetadata()

    folder_reader_messages: Annotated[
        List[GenericMessage], Field(description="Folder reader messages.")
    ] = []
    db_reader_messages: Annotated[
        List[GenericMessage], Field(description="DB reader messages.")
    ] = []
    has_db_metadata: Annotated[
        bool, Field(description="True if database metadata is loaded.")
    ] = False
    has_sample_table_data: Annotated[
        bool, Field(description="True if sample file is loaded.")
    ] = False
    has_assay_table_data: Annotated[
        bool, Field(description="True if assay files are loaded.")
    ] = False
    has_assignment_table_data: Annotated[
        bool, Field(description="True if assignment files are loaded")
    ] = False
    has_folder_metadata: Annotated[
        bool, Field(description="True if folder metadata is complete.")
    ] = False
    has_investigation_data: Annotated[
        bool, Field(description="True if investtigation file is loaded.")
    ] = False



from pydantic import Field
from typing_extensions import Annotated

from metabolights_utils.common import CamelCaseModel
from metabolights_utils.models.enums import GenericMessageType


class MetabolightsBaseModel(CamelCaseModel):
    pass


class GenericMessage(MetabolightsBaseModel):
    type: Annotated[GenericMessageType, Field(description="Message type.")] = (
        GenericMessageType.ERROR
    )
    short: Annotated[str, Field(description="")] = ""
    detail: Annotated[str, Field(description="")] = ""

    def __str__(self) -> str:
        value = f"{self.type}\tshort: {self.short}\tdetail: {self.detail}"
        return value


class InfoMessage(GenericMessage):
    type: Annotated[
        GenericMessageType, Field(description="Information message type.")
    ] = GenericMessageType.INFO


class WarningMessage(GenericMessage):
    type: Annotated[GenericMessageType, Field(description="Warning message type.")] = (
        GenericMessageType.WARNING
    )


class CriticalMessage(GenericMessage):
    type: Annotated[GenericMessageType, Field(description="Critical message type.")] = (
        GenericMessageType.CRITICAL
    )


class ErrorMessage(GenericMessage):
    type: Annotated[GenericMessageType, Field(description="Error message type")] = (
        GenericMessageType.ERROR
    )



import os
from typing import Any, Dict, Union

import click
from pydantic.alias_generators import to_camel, to_snake

from metabolights_utils.commands.utils import split_to_lines
from metabolights_utils.models.metabolights.model import MetabolightsStudyModel


@click.command(name="explain")
@click.argument("model_pattern", required=False)
def model_explain(
    model_pattern: Union[None, str] = None,
):
    """Expain properties and sub-properties of MetaboLights study model. It lists root properties of the model, If it is not specified

    Examples:
    investigation -> explains properties of Investigation property of MetaboLights study model.
    investigation.studies -> explains properties of Study model
    investigation.studies.study_assays.assays -> explains properties of Assay model
    investigation.studies.study_assays.comments -> explains properties of Comment model of assay

    assays.assay_technique -> explains Assay Technique model

    """
    search_pattern = model_pattern
    schema = MetabolightsStudyModel.model_json_schema(
        by_alias=True, mode="serialization"
    )
    model_name = None
    title = None
    description = None
    if not search_pattern:
        model = schema
        model_name = "object"
        title = "MetaboLights Study Model"
        description = ""
    else:
        model, model_name, title, description = find_property(schema, search_pattern)

    click.echo(f"\nExplain '{search_pattern}'")
    if title:
        print_properties(schema, model_name, title, description, model)
    else:
        click.echo(f"Not valid search object '{search_pattern}'")


def print_properties(
    schema: Dict[str, Any],
    model_name: str,
    title: str,
    description: str,
    model: Dict[str, Any],
):
    required_properties = set()
    if model and "required" in model:
        required_properties = {x for x in model["required"]}
    properties = {}
    _type = None
    if model and "properties" in model:
        _type = define_type(schema, model["properties"])
        properties = model["properties"]
    enums = ""
    click.echo(f"\nDefinition: {title} [{model_name}]")

    if model and "enum" in model:
        enum = f", ".join([str(x) for x in model["enum"]])
        enums = "Valid values: " + "\n\t ".join(split_to_lines(enum))
    if description:
        dec = split_to_lines(description)
        click.echo("\t" + f"\n\t".join(dec) + (f" {enums}" if enums else ""))
    elif enums:
        click.echo(f"\t{enums}")
    property_names = list([x for x in properties.keys()])
    if property_names:
        property_names.sort()

        click.echo("\nProperties:")
        for property in property_names:
            target = properties[property]
            required_properties = set()
            if "" in target:
                required_properties = {x for x in target["required"]}

            property_name = to_snake(property)
            if "isatab_config" in property_name:
                continue
            required = f"{'***required***' if property in required_properties else ''}"

            _type = define_type(schema, target)

            click.echo(
                "  " + property_name + f" [{_type}] {required if required else ''}"
            )
            if "description" in target:
                dec = split_to_lines(target["description"])

            click.echo("\t" + f"\n\t".join(dec).strip() + "\n")


def define_type(schema: Dict[str, Any], desc: Dict[str, Any]):
    _type = ""
    if "type" in desc:
        _type = desc["type"]
    if not _type:
        if "allOf" in desc:
            types = []
            for item in desc["allOf"]:
                definition = item
                if "$ref" in definition:
                    types.append(definition["$ref"].split("/")[-1])
                else:
                    if "type" in item:
                        types.append(item["type"])
            _type = ", ".join(types)
            if len(types) > 1:
                _type = f"All of {_type}"
        elif "anyOf" in desc:
            types = []
            for item in desc["anyOf"]:
                definition = item
                if "$ref" in definition:
                    types.append(definition["$ref"].split("/")[-1])
                else:
                    if "type" in item:
                        types.append(item["type"])
            _type = ", ".join(types)
            if len(types) > 1:
                _type = f"Any of {_type}"

    else:

        if _type == "object" and "properties" not in desc:
            _type = "dictionary of"
        elif _type == "array":
            _type = "array of"

        if "additionalProperties" in desc and isinstance(
            desc["additionalProperties"], dict
        ):
            additional = desc["additionalProperties"]
            types = []
            if "$ref" in additional:
                child = additional["$ref"].split("/")[-1]
                if child in schema["$defs"]:
                    _type = f"{_type} {child}".strip()
                else:
                    sub_type = define_type(schema, schema["$defs"][child])
                    _type = f"{_type} {sub_type}".strip()
            else:
                child = None
                prefix = ""
                if "items" in additional:
                    items = additional["items"]
                    child_name = ""
                    if "$ref" in items:
                        child_name = items["$ref"].split("/")[-1]
                    if child_name in schema["$defs"]:
                        items = schema["$defs"][child_name]
                    if "properties" in items:
                        if "type" in items and items["type"] == "object":
                            return f"array of {child_name}"
                        else:
                            child = define_type(schema, items)
                    elif "type" in items:
                        child = items["type"]

                    prefix = "array of"
                    _type = f"{_type} {prefix} {child}".strip()
                elif "$ref" in additional:

                    child_name = additional["$ref"].split("/")[-1]
                    child = define_type(schema, schema["$defs"][child_name])
                    prefix = "dictionary of"
                    _type = f"{_type} {prefix} {child}".strip()

        elif "items" in desc:
            if "$ref" in desc["items"]:
                child = ""
                child = desc["items"]["$ref"].split("/")[-1]
            else:
                if "type" in desc["items"]:
                    child = desc["items"]["type"]
            if child:
                # if child in schema["$defs"]:
                #     _type = f"dictionary of {child}"
                # else:
                _type = f"array of {child}"

    return _type


def find_property(schema: Dict[str, Any], search_str: str):
    if "properties" not in schema:
        return None, None, None, None
    values = search_str.split(".")
    if not values:
        return None, None, None, None
    current_description = None
    current_model = schema

    current_properties = []
    if to_camel(search_str) in schema["$defs"]:
        current_model = schema["$defs"][to_camel(search_str)]
    if "properties" in current_model:
        current_properties = current_model["properties"]

    model_title_terms = []
    model_name = None
    not_matched = False
    if not values or "properties" not in current_model:
        return None, None, None, None
    for value in values:
        found = False
        property_name = to_camel(value)
        if "isatab_config" in property_name:
            continue
        if property_name in current_properties:

            definition = current_properties[property_name]
            _type = define_type(schema, definition)
            if _type and _type.split()[-1]:
                model = _type.split()[-1]
                if model in schema["$defs"]:
                    if "properties" in schema["$defs"][model]:
                        current_model = schema["$defs"][model]
                        current_properties = current_model["properties"]
                    elif "enum" in schema["$defs"][model]:
                        current_model = schema["$defs"][model]
                        current_properties = schema["$defs"][model]
                    else:
                        current_model = None
                        current_properties = None

                else:
                    current_model = None
                    current_properties = None

                if "description" in definition:
                    current_description = definition["description"]
                # if current_model and "enum" in current_model:
                #     enum = f", ".join([str(x) for x in current_model["enum"]])
                #     enums = "Valid values: " + "\n\t ".join(split_to_lines(enum))
                #     if current_description:
                #         current_description += f" {enums}"

                model_name = _type if _type else to_snake(property_name)
                model_title_terms.append(to_snake(property_name))

                found = True

            if not found:
                not_matched = True
                break
        else:
            not_matched = True
            break
    if not_matched:
        return None, None, None, current_description
    return (
        current_model,
        model_name,
        " -> ".join(model_title_terms),
        current_description,
    )


if __name__ == "__main__":
    model_explain()



import glob
import hashlib
import os
from typing import Dict

from pydantic import BaseModel

from metabolights_utils.utils.search_utils import MetabolightsSearchUtils as SearchUtils

EMPTY_FILE_HASH = hashlib.sha256("".encode()).hexdigest()


class IsaMetadataFolderHash(BaseModel):
    folder_sha256: str = ""
    files_sha256: Dict[str, str] = {}


class MetabolightsHashUtils(object):
    @staticmethod
    def sha256sum(filepath: str):
        if not filepath or not os.path.exists(filepath):
            return EMPTY_FILE_HASH

        sha256_hash = hashlib.sha256()
        with open(filepath, "rb") as f:
            for byte_block in iter(lambda: f.read(4096), b""):
                sha256_hash.update(byte_block)
        return sha256_hash.hexdigest()

    @staticmethod
    def get_isa_metadata_folder_hash(folder_path: str) -> IsaMetadataFolderHash:
        files: list[str] = SearchUtils.get_isa_metadata_files(
            folder_path=folder_path, recursive=False
        )
        files.sort()
        result = IsaMetadataFolderHash()
        hashes = []
        for file in files:
            basename = os.path.basename(file)
            hash = MetabolightsHashUtils.sha256sum(file)
            result.files_sha256[basename] = hash
            hashes.append(f"{basename}:{hash}")
        hash_bytes = ",".join(hashes).encode("utf-8")
        result.folder_sha256 = hashlib.sha256(hash_bytes).hexdigest()
        return result



import os
import re
import shutil
from ftplib import FTP, error_perm
from typing import List, Set, Union

from metabolights_utils.provider.ftp.model import FtpFolderContent, LocalDirectory

code_pattern = re.compile(r"\s*(\d+)\s+(.*)\s*")


class DefaultFtpClient:
    def __init__(
        self,
        local_storage_root_path: str,
        ftp_server_url: str,
        remote_repository_root_directory: str,
        username: Union[str, None] = None,
        password: Union[str, None] = None,
    ) -> None:
        self.ftp_server_url = ftp_server_url
        self.remote_repository_root_directory = remote_repository_root_directory
        self.local_storage_root_path = local_storage_root_path
        self.username = username if username else ""
        self.password = password if password else ""
        os.makedirs(self.local_storage_root_path, exist_ok=True)

        self.ftp: Union[None, FTP] = None

    def is_ftp_directory(self, name):
        current = self.ftp.pwd()
        try:
            self.ftp.cwd(name)
            self.ftp.cwd(current)
            return True
        except Exception:
            return False

    def connect(self):
        if not self.ftp:
            self.ftp = FTP(
                self.ftp_server_url,
                timeout=10.0,
                user=self.username,
                passwd=self.password,
            )
        self.ftp.login(user=self.username, passwd=self.password)

    def quit(self):
        if self.ftp:
            try:
                self.ftp.quit()
            except Exception as ex:
                print(str(ex))
            self.ftp = None

    def list_directory(
        self,
        directory: Union[str, None] = None,
        search_pattern: Union[str, None] = None,
    ) -> FtpFolderContent:
        directory = directory.strip("/") if directory else ""
        root_dir = self.remote_repository_root_directory
        remote_directory = os.path.join(root_dir, directory)

        self.connect()

        command = f"LIST {search_pattern}" if search_pattern else "LIST"
        response: FtpFolderContent = FtpFolderContent(source_directory=remote_directory)
        try:
            self.ftp.cwd(remote_directory)
            return_code = self.ftp.retrlines(command, callback=response.parse_line)
            result = code_pattern.search(return_code)
            if result and result.groups():
                response.code = int(result.groups()[0])
                response.message = result.groups()[1]
            response.success = True
            return response
        except Exception as ex:
            result = code_pattern.search(str(ex))
            if result and result.groups():
                response.code = int(result.groups()[0])
                response.message = result.groups()[1]
            return response
        finally:
            self.quit()

    def upload_files(
        self,
        remote_directory: Union[str, None] = None,
        file_paths: Union[List[str], None] = None,
    ) -> None:
        remote_directory = remote_directory.strip("/") if remote_directory else ""
        self.connect()

        try:
            self.ftp.cwd(remote_directory)
            for file_path in file_paths:
                filename = os.path.basename(file_path)

                with open(file_path, "rb") as file:
                    self.ftp.storbinary(f"STOR {filename}", file)
        finally:
            self.quit()

    def download_file(
        self,
        relative_file_path: str,
        local_path: Union[List[str], None] = None,
        override_local_files: bool = False,
        local_files: Union[LocalDirectory, None] = None,
        skip_files: Union[Set[str], None] = None,
        delete_unlisted_local_files: bool = True,
        keep_local_files: Union[Set[str], None] = None,
    ) -> LocalDirectory:
        local_path = local_path if local_path else self.local_storage_root_path
        if local_files is None:
            response = LocalDirectory(root_path=local_path)
        else:
            local_files.root_path = local_path
            response = local_files
        actions = response.actions

        if not relative_file_path or not relative_file_path.strip():
            return LocalDirectory(
                root_path=local_path,
                code=451,
                message="root path download does not allowed.",
            )
        if skip_files and relative_file_path in skip_files:
            return response
        remote_root_dir = self.remote_repository_root_directory

        target_path = os.path.join(local_path, relative_file_path)
        parent_path = os.path.dirname(target_path)
        filename = os.path.basename(target_path)

        remote_directory = os.path.join(remote_root_dir, relative_file_path)
        relative_parent_path = os.path.dirname(relative_file_path)

        remote_parent_directory = os.path.dirname(remote_directory)
        result = self.list_directory(relative_parent_path, search_pattern=filename)
        is_directory = True
        target_exists = False
        if result.success:
            if (
                result.descriptors
                and len(result.descriptors) == 1
                and result.descriptors[0].base_name == filename
            ):
                is_directory = False
            target_exists = True
        if not target_exists:
            return LocalDirectory(
                root_path=local_path, code=result.code, message=result.message
            )
        try:
            if not is_directory:
                self.connect()
                self.ftp.cwd(remote_parent_directory)

                if not os.path.exists(target_path) or override_local_files:
                    os.makedirs(parent_path, exist_ok=True)

                    with open(target_path, "wb") as local_file:
                        self.ftp.retrbinary("RETR " + filename, local_file.write)
                    actions[relative_file_path] = "DOWNLOADED"
                else:
                    actions[relative_file_path] = "SKIPPED"
                response.local_files.append(relative_file_path)
            else:

                if not os.path.exists(target_path):
                    actions[relative_file_path] = "FOLDER_CREATED"
                    os.makedirs(target_path, exist_ok=True)
                else:
                    actions[relative_file_path] = "FOLDER_SKIPPED"
                    if delete_unlisted_local_files:
                        local_file_list = os.listdir(target_path)
                        listed_folders = {
                            x.base_name for x in result.descriptors if x.is_directory
                        }
                        listed_files = {
                            x.base_name
                            for x in result.descriptors
                            if not x.is_directory
                        }
                        for item in local_file_list:
                            if item not in listed_folders and item not in listed_files:
                                item_path: str = os.path.join(target_path, item)
                                relative_item_path = item_path.replace(
                                    f"{local_path}", ""
                                ).strip("/")
                                if keep_local_files and item in keep_local_files:
                                    actions[relative_item_path] = "SKIPPED"
                                    continue
                                try:
                                    if os.path.isdir(item_path):
                                        shutil.rmtree(item_path)
                                        actions[relative_item_path] = "FOLDER_DELETED"
                                    else:
                                        os.remove(item_path)
                                        actions[relative_item_path] = "DELETED"
                                except Exception as ex:
                                    print(str(ex))
                response.local_folders.append(relative_file_path)
                for collection in (result.folders, result.files):
                    for entry in collection:
                        new_relative_file_path = os.path.join(relative_file_path, entry)
                        # new_local_path = os.path.join(local_path, entry)
                        self.download_file(
                            relative_file_path=new_relative_file_path,
                            local_path=local_path,
                            override_local_files=override_local_files,
                            local_files=response,
                            skip_files=skip_files,
                        )

            response.success = True
            response.code = 200
            response.message = "Ok"
        except Exception as ex:
            print(ex)
            response.message = str(ex)
            response.code = 400
        finally:
            self.quit()

        return response



import io
import re
from typing import Callable, List

from bs4 import BeautifulSoup

from metabolights_utils.models.enums import GenericMessageType
from metabolights_utils.models.isa.common import IsaTable
from metabolights_utils.models.metabolights.model import MetabolightsStudyModel
from metabolights_utils.models.parser.enums import ParserMessageType


def sub_arrays(arr: List[str], k: int):
    arr.sort()
    count = 0
    array_list = []
    current_list = None
    for item in arr:
        if count % k == 0:
            current_list = []
            array_list.append(current_list)
        current_list.append(item)
        count += 1
    return array_list
    # return set(chain.from_iterable(combinations(arr, r) for r in range(k + 1)))


def start_section(message: str, length: int = 120):
    print("\n")
    print("-" * length)
    if message:
        print(message)


def print_list(files: List[str], remove_prefix: str):
    files.sort(key=lambda x: sort_filename(x, remove_prefix=remove_prefix))
    for file in files:
        print(f"\t{file}")


def sort_filename(filename: str, remove_prefix: str = None):
    name = filename
    if remove_prefix:
        name = filename.removeprefix(remove_prefix).strip("/")

    name = filename.removeprefix(remove_prefix).strip("/")
    if name.startswith("FILES/"):
        return f"_{name}"
    return name


def find_column_names(isa_table: IsaTable, pattern: str):
    column_names = []
    for column in isa_table.columns:
        if re.match(pattern, column):
            column_names.append(column)
    return column_names


def get_unique_values(values: List[str]) -> List[str]:
    if not values:
        return []
    value_set = set([x.strip() for x in values if x.strip()])
    unique_values = list(value_set)
    unique_values.sort()
    return unique_values


def convert_html_to_plain_text(text: str):
    f = io.BytesIO(bytes(text, encoding="utf-8"))
    soup = BeautifulSoup(f, "html.parser")
    return soup.text


def split_to_lines(text: str, max_line_size=120, sep=None, join_term=" "):
    join_term_len = len(join_term)
    if sep:
        words = text.split(sep)
    else:
        words = text.split()
    count = 0
    array_list = []
    current_list = []
    for word in words:
        if count + len(word) + join_term_len > max_line_size:
            array_list.append(" ".join(current_list))
            current_list = []
            count = 0
        current_list.append(word)
        count += len(word) + join_term_len if count > 0 else len(word)
    if current_list:
        array_list.append(join_term.join(current_list))
    return array_list


assay_column_search_patterns = [
    r".*Parameter.+Value.*\[.+\].*",
    r".*Metabolite.*Assignment.*File.*",
]


def print_study_model_summary(model: MetabolightsStudyModel, log: Callable = print):
    study = model.investigation.studies[0]
    study_id = study.identifier
    log(f"{study.identifier}: {study.title}")
    public_date = study.public_release_date
    submission_date = study.submission_date
    contacts = study.study_contacts.people
    lines = split_to_lines(convert_html_to_plain_text(study.description))
    for line in lines:
        log(f"\t{line}")
    log(f"\nRelease Date\t: {public_date}\nSubmission Date\t: {submission_date}")
    submitters = ", ".join([f"{x.first_name} {x.last_name}" for x in contacts])
    log(f"Submitter(s)\t: {submitters}\n")
    descriptors = study.study_design_descriptors.design_types
    descriptor_format = "\n".join([f"\t- {str(x)}" for x in descriptors])
    log(f"\nDescriptors:\n{descriptor_format}")
    factors = study.study_factors.factors
    factors_format = "\n".join([f"\t- {str(x)}" for x in factors])
    log(f"Factors:\n{factors_format}")

    protocols = study.study_protocols.protocols
    log(f"Protocols:")
    for protocol in protocols:
        lines = split_to_lines(convert_html_to_plain_text(protocol.description))
        log(f"\t- {protocol.name}")
        for line in lines:
            log(f"\t\t{line}")
    # protocol_format = "\n".join(
    #     [f"\t- {x.name}\n\t\t{convert_html_to_plain_text(x.description)}" for x in protocols]
    # )
    # log(f"Protocols:\n{protocol_format}")

    if model.assays:
        log("\nAssays:")
    for assay_filename, assay in model.assays.items():
        report: List[str] = []
        report.append(f"- {assay.file_path}")
        report.append(
            f"Row count: {assay.table.total_row_count} Column Count: {assay.table.total_column_count}"
        )
        report.append(f"Technique: {assay.assay_technique}")
        report.append(f"Raw Files: {len(assay.referenced_raw_files)}")
        raw_extensions = assay.referenced_raw_file_extensions
        report.append(f"Raw File Extensions: {', '.join(raw_extensions)}")
        report.append(f"Derived Files: {len(assay.referenced_derived_files)}")
        derived_extensions = assay.referenced_derived_file_extensions
        report.append(f"Derived File Extensions: {', '.join(derived_extensions)}")
        assignments = assay.referenced_assignment_files
        print_columns(report, assay.table, assay_column_search_patterns, indent_tab=3)
        stats = "\n\t\t- ".join(report)
        log(f"\t{stats}")

    if model.metabolite_assignments:
        log("\nMetabolite Assignment Files (MAF):")
    for maf_filename, maf in model.metabolite_assignments.items():
        report: List[str] = []
        report.append(f"- {maf.file_path}")
        report.append(f"Row count: {maf.number_of_rows}")
        report.append(f"Main technique: {maf.assay_technique.main_technique}")
        report.append(f"Annotated metabolites: {len(maf.metabolite_assignments)}")
        stats = "\n\t\t- ".join(report)
        log(f"\t{stats}")

        if maf.metabolite_assignments:
            log("\t\t- Metabolite Annotations:")
            assignments = [
                f"{maf.metabolite_assignments[x]}:{x}"
                for x in maf.metabolite_assignments
            ]
            lines = split_to_lines(
                convert_html_to_plain_text(", ".join(assignments)), sep=", "
            )
            for line in lines:
                log(f"\t\t\t{line}")
        else:
            log("\t\t- Metabolite Annotations: - ")

    log("\nSample File:")
    sample_filepath = study.file_name
    sample_file = model.samples[sample_filepath]
    report = []
    patterns = [r".*Characteristics.*\[.+\].*", r".*Factor.+Value.*\[.+\].*"]
    print_columns(report, sample_file.table, patterns, indent_tab=2)
    stats = "\n\t- ".join(report)
    log(f"\t- {stats}")
    log("")

    errors = []
    for file, messages in model.parser_messages.items():
        for message in messages:
            if message.type in (ParserMessageType.CRITICAL, ParserMessageType.ERROR):
                errors.append(f"{study_id} {file}: {message.short}")
    for message in model.db_reader_messages:
        if message.type in (GenericMessageType.CRITICAL, GenericMessageType.ERROR):
            errors.append(f"{study_id} DB: {message.short}\t{message.detail}")
    for message in model.folder_reader_messages:
        if message.type in (GenericMessageType.CRITICAL, GenericMessageType.ERROR):
            errors.append(f"{study_id} Folder: {message.short}\t{message.detail}")
    if errors:
        log("Model Errors:")
        for error in errors:
            lines = split_to_lines(error)
            for line in lines:
                log(f"\t- {line}")


def print_columns(
    report: List[str],
    isa_table: IsaTable,
    patterns: List[str],
    indent_tab=2,
    max_item_in_row=6,
):
    tab = "\t" * indent_tab
    for pattern in patterns:
        column_names = find_column_names(isa_table, pattern)
        for colum_name in column_names:
            unique_values = get_unique_values(isa_table.data[colum_name])
            if unique_values:
                if len(unique_values) < 6:
                    report.append(f"{colum_name}: {', '.join(unique_values)}")
                else:
                    arrays = sub_arrays(unique_values, max_item_in_row)
                    terms = []
                    for array in arrays:
                        terms.append(f"{', '.join(array)}")
                    joined_terms = "- " + f"\n{tab}- ".join(terms)
                    report.append(f"{colum_name}:\n{tab}{joined_terms}")
            else:
                report.append(f"{colum_name}: -")



import os
import pathlib
from abc import ABC, abstractmethod
from io import IOBase
from typing import List, Union

from metabolights_utils.isatab.default.base_isa_file import BaseIsaFile
from metabolights_utils.isatab.default.parser.isa_table_parser import get_isa_table_file
from metabolights_utils.isatab.reader import (
    IsaTableFileReader,
    IsaTableFileReaderResult,
)
from metabolights_utils.models.isa.common import IsaTable, IsaTableFile
from metabolights_utils.models.parser.common import ParserMessage, ParserReport
from metabolights_utils.models.parser.enums import ParserMessageType
from metabolights_utils.tsv.filter import TsvFileFilterOption
from metabolights_utils.tsv.sort import TsvFileSortOption
from metabolights_utils.utils.hash_utils import MetabolightsHashUtils as HashUtils


class BaseIsaTableFileReader(BaseIsaFile, IsaTableFileReader, ABC):
    def __init__(self, results_per_page=100) -> None:
        self.results_per_page = results_per_page if results_per_page > 0 else 100

    @abstractmethod
    def get_expected_patterns(self) -> List[List[str]]:
        pass

    def get_total_pages(
        self,
        file_buffer_or_path: Union[str, pathlib.Path, IOBase],
        results_per_page: int = 100,
    ) -> int:
        total = self.get_total_row_count(file_buffer_or_path)
        return int(total / results_per_page) + (
            1 if total % results_per_page > 0 else 0
        )

    def get_total_row_count(
        self,
        file_buffer_or_path: Union[str, pathlib.Path, IOBase],
    ) -> int:
        isa_file_result: IsaTableFileReaderResult = self.get_headers(
            file_buffer_or_path
        )
        table_file: IsaTableFile = isa_file_result.isa_table_file
        table: IsaTable = table_file.table
        return table.total_row_count

    def get_page(
        self,
        file_buffer_or_path: Union[str, pathlib.Path, IOBase],
        page: int = 1,
        results_per_page: int = 100,
        selected_columns: Union[List[str], None] = None,
        filter_options: List[TsvFileFilterOption] = None,
        sort_options: List[TsvFileSortOption] = None,
    ) -> IsaTableFileReaderResult:
        page = page if page and page > 1 else 1
        results_per_page = (
            results_per_page if results_per_page > 0 else self.results_per_page
        )
        offset = (page - 1) * results_per_page
        return self.get_rows(
            file_buffer_or_path=file_buffer_or_path,
            offset=offset,
            limit=results_per_page,
            selected_columns=selected_columns,
            filter_options=filter_options,
            sort_options=sort_options,
        )

    def get_rows(
        self,
        file_buffer_or_path: Union[str, pathlib.Path, IOBase],
        offset: int = 0,
        limit: Union[int, None] = None,
        selected_columns: Union[None, List[str]] = None,
        filter_options: List[TsvFileFilterOption] = None,
        sort_options: List[TsvFileSortOption] = None,
    ) -> IsaTableFileReaderResult:
        offset = offset if offset > 0 else None
        limit = limit if limit is not None and limit >= 0 else None
        selected_columns = selected_columns if selected_columns else None

        return self.read(
            file_buffer_or_path=file_buffer_or_path,
            offset=offset,
            limit=limit,
            selected_columns=selected_columns,
            filter_options=filter_options,
            sort_options=sort_options,
        )

    def get_headers(
        self,
        file_buffer_or_path: Union[str, pathlib.Path, IOBase],
    ) -> IsaTableFileReaderResult:
        offset = None
        limit = 0
        selected_columns = None

        return self.read(
            file_buffer_or_path=file_buffer_or_path,
            offset=offset,
            limit=limit,
            selected_columns=selected_columns,
            skip_parser_info_messages=True,
        )

    def read(
        self,
        file_buffer_or_path: Union[str, pathlib.Path, IOBase],
        offset: Union[None, int],
        limit: Union[None, int],
        selected_columns: Union[None, List[str]] = None,
        skip_parser_info_messages: bool = True,
        filter_options: List[TsvFileFilterOption] = None,
        sort_options: List[TsvFileSortOption] = None,
    ) -> IsaTableFileReaderResult:
        buffer: IOBase = None
        read_messages: List[ParserMessage] = []
        buffer_or_path, path = self._get_file_path(file_buffer_or_path)
        basename = os.path.basename(str(path))
        isa_table_file = None
        try:
            file_buffer = self._get_file_buffer(buffer_or_path)
            isa_table_file: IsaTableFile = get_isa_table_file(
                file_buffer,
                basename,
                messages=read_messages,
                expected_patterns=self.get_expected_patterns(),
                offset=offset,
                limit=limit,
                selected_columns=selected_columns,
                filter_options=filter_options,
                sort_options=sort_options,
            )
            messages = read_messages
        except UnicodeDecodeError as err:
            try:
                file_buffer = self._get_file_buffer(buffer_or_path, encoding="ascii")
                isa_table_file: IsaTableFile = get_isa_table_file(
                    file_buffer,
                    basename,
                    messages=read_messages,
                    expected_patterns=self.get_expected_patterns(),
                    offset=offset,
                    limit=limit,
                    selected_columns=selected_columns,
                    filter_options=filter_options,
                    sort_options=sort_options,
                )
                messages = read_messages
            except Exception as exc:
                read_messages.append(
                    ParserMessage(
                        type=ParserMessageType.CRITICAL,
                        short="Parse error",
                        detail=(str(exc)),
                    )
                )
        except Exception as exc:
            read_messages.append(
                ParserMessage(
                    type=ParserMessageType.CRITICAL,
                    short="Parse error",
                    detail=(str(exc)),
                )
            )
        finally:
            self._close_file(file_buffer_or_path)
        if isa_table_file:
            if os.path.exists(path):
                isa_table_file.sha256_hash = HashUtils.sha256sum(path)
            elif os.path.exists(str(file_buffer_or_path)):
                isa_table_file.sha256_hash = HashUtils.sha256sum(
                    str(file_buffer_or_path)
                )
        else:
            isa_table_file = IsaTableFile()

        if skip_parser_info_messages:
            messages = [x for x in read_messages if x.type != ParserMessageType.INFO]
        parser_report = ParserReport(messages=messages)
        return IsaTableFileReaderResult(
            isa_table_file=isa_table_file, parser_report=parser_report
        )



import sys
from abc import ABC, abstractmethod
from datetime import datetime
from enum import Enum, EnumMeta
from typing import Any, Callable, Dict, List, Tuple

from pydantic import Field
from typing_extensions import Annotated

from metabolights_utils.common import CamelCaseModel


class SortType(str, Enum):
    STRING = "string"
    INTEGER = "integer"
    FLOAT = "float"
    DATETIME = "datetime"
    CUSTOM = "custom"


class SortValueClassification(int, Enum):
    EMPTY = 1
    INVALID = 2
    VALID = 3


class TsvFileSortValueOrder(int, Enum):
    EMPTY_INVALID_VALID = 0o123
    EMPTY_VALID_INVALID = 0o132
    INVALID_EMPTY_VALID = 0o213
    INVALID_VALID_EMPTY = 0o231
    VALID_INVALID_EMPTY = 0o321
    VALID_EMPTY_INVALID = 0o312


class TsvFileSortOption(CamelCaseModel):
    column_name: Annotated[
        str,
        Field(
            description="Column name that sort operation will be applied."
            + "Column name may be different if there are multiple columns with same name."
            "If there are multiple column headers, Second column and others are named as <column header>.1 <column header>.2 and so on.",
        ),
    ] = ""

    reverse: Annotated[bool, Field(description="Sort descending order")] = False

    column_sort_type: Annotated[
        SortType, Field(description="Sort value with the given type")
    ] = SortType.STRING

    case_sensitive: Annotated[
        bool, Field(description="Sort string values in case sensitive order")
    ] = True

    default_datetime_pattern: Annotated[
        str, Field(description="Convert string value to datetime using this pattern")
    ] = "%d/%m/%Y"

    value_order: Annotated[
        TsvFileSortValueOrder,
        Field(
            description="Position of valid and empty values if sort type is int, float or datetime",
        ),
    ] = TsvFileSortValueOrder.VALID_EMPTY_INVALID

    custom_sorter_name: Annotated[
        str, Field(description="Registered custom sorter name")
    ] = ""

    custom_sorter_arguments: Annotated[
        Dict[str, Any], Field(description="Additional arguments for custom sorters")
    ] = {}

    min_key_length: Annotated[
        int,
        Field(
            description="If sort type is not string, values will be converted to string and short strings will be filled with 0.",
        ),
    ] = -1


class TsvSortException(Exception):
    def __init__(self, message: str="") -> None:
        super().__init__(message)


class Sorter(ABC):
    def __init__(
        self,
        sort_option: TsvFileSortOption,
        column_idx: int,
        column_name_indices: Dict[str, int],
        column_indices: Dict[int, str],
    ) -> None:
        self.sort_option = sort_option
        self.column_idx = column_idx
        self.column_name_indices = column_name_indices
        self.column_indices = column_indices

        if not sort_option or not sort_option.column_name:
            raise TsvSortException("Invalid sort option")

        if self.column_idx not in self.column_indices:
            raise TsvSortException("Invalid sort index")

        if sort_option.column_name not in self.column_name_indices:
            raise TsvSortException("Column name is not in table columns")

    @abstractmethod
    def sort(self, row: Tuple[int, List[str]]) -> str:
        pass


class SorterRegistry:
    custom_sorters: Dict[str, Callable] = {}
    default_sorters: Dict[str, Callable] = {}

    @classmethod
    def register_sorter(cls, name: str, sorter: Callable) -> None:
        cls.default_sorters[name] = sorter

    @classmethod
    def get_sorter(
        cls,
        sort_option: TsvFileSortOption,
        column_idx: int,
        column_name_indices: Dict[str, int],
        column_indices: Dict[int, str],
    ) -> Sorter:
        name = ""
        if sort_option.column_sort_type == SortType.CUSTOM:
            name = sort_option.custom_sorter_name
            if name in cls.custom_sorters:
                return cls.custom_sorters[name](
                    sort_option, column_idx, column_name_indices, column_indices
                )
        else:
            name = sort_option.column_sort_type.value
            if name in cls.default_sorters:
                return cls.default_sorters[name](
                    sort_option, column_idx, column_name_indices, column_indices
                )
        raise TsvSortException(f"Sorter name {name} is not registered")

    @classmethod
    def register_custom_sorter(cls, name: str, sorter: Any) -> None:
        cls.custom_sorters[name] = sorter

    @classmethod
    def unregister_custom_sorter(cls, name: str) -> None:
        if name in cls.custom_sorters:
            del cls.custom_sorters[name]


# SORTERS


class AbstractSorter(Sorter, ABC):
    NONE_VALUE = "".rjust(20, "0")

    def __init__(
        self,
        sort_option: TsvFileSortOption,
        column_idx: int,
        column_name_indices: Dict[str, int],
        column_indices: Dict[int, str],
        default_min_key_length: int = 20,
    ) -> None:
        super().__init__(sort_option, column_idx, column_name_indices, column_indices)
        self.min_key_length = sort_option.min_key_length
        if self.min_key_length < 0:
            self.min_key_length = default_min_key_length

        self.invalid_value_order = self.get_sort_value_order(
            SortValueClassification.INVALID
        )
        self.none_value_order = self.get_sort_value_order(SortValueClassification.EMPTY)
        self.valid_value_order = self.get_sort_value_order(
            SortValueClassification.VALID
        )

    def get_invalid_value(self, value: str):
        length = self.min_key_length
        if self.sort_option.case_sensitive:
            return str(self.invalid_value_order) + str(value).zfill(length)
        else:
            return str(self.invalid_value_order) + str(value.lower()).zfill(length)

    def get_none_value(self):
        return str(self.none_value_order) + self.NONE_VALUE

    def get_sort_value_order(self, classification: SortValueClassification):
        if self.sort_option.value_order & 0o007 == classification:
            return 1 if self.sort_option.reverse else 3
        elif (self.sort_option.value_order & 0o070) >> 3 == classification:
            return 2
        return 3 if self.sort_option.reverse else 1

    def sort(self, row: Tuple[int, List[str]]) -> str:
        value = row[1][self.column_idx]
        if not value:
            return self.get_none_value()
        try:
            order = self.valid_value_order
            length = self.min_key_length
            if length > 0:
                return str(order) + self.get_sorted_string(value).zfill(length)
            else:
                return str(order) + self.get_sorted_string(value)
        except Exception:
            return self.get_invalid_value(value)

    @abstractmethod
    def get_sorted_string(self, value: str) -> str:
        pass


class CustomSorter(AbstractSorter, ABC):
    def __init__(
        self,
        sort_option: TsvFileSortOption,
        column_idx: int,
        column_name_indices: Dict[str, int],
        column_indices: Dict[int, str],
        default_min_key_length: int = 0,
    ) -> None:
        super().__init__(
            sort_option,
            column_idx,
            column_name_indices,
            column_indices,
            default_min_key_length=default_min_key_length,
        )


class IntegerSorter(AbstractSorter):
    def __init__(
        self,
        sort_option: TsvFileSortOption,
        column_idx: int,
        column_name_indices: Dict[str, int],
        column_indices: Dict[int, str],
    ) -> None:
        super().__init__(sort_option, column_idx, column_name_indices, column_indices)

    def get_sorted_string(self, value: str) -> str:
        return str(int(value))


class FloatSorter(AbstractSorter):
    def __init__(
        self,
        sort_option: TsvFileSortOption,
        column_idx: int,
        column_name_indices: Dict[str, int],
        column_indices: Dict[int, str],
    ) -> None:
        super().__init__(sort_option, column_idx, column_name_indices, column_indices)

    def get_sorted_string(self, value: str) -> float:
        return str(int(float(value) * 100000))


class DateTimeSorter(AbstractSorter):
    def __init__(
        self,
        sort_option: TsvFileSortOption,
        column_idx: int,
        column_name_indices: Dict[str, int],
        column_indices: Dict[int, str],
    ) -> None:
        super().__init__(sort_option, column_idx, column_name_indices, column_indices)

    def get_sorted_string(self, value: str) -> datetime:
        val = datetime.strptime(value, self.sort_option.default_datetime_pattern)
        return str(int(val.timestamp()))


class StringSorter(AbstractSorter):
    def __init__(
        self,
        sort_option: TsvFileSortOption,
        column_idx: int,
        column_name_indices: Dict[str, int],
        column_indices: Dict[int, str],
    ) -> None:
        super().__init__(
            sort_option,
            column_idx,
            column_name_indices,
            column_indices,
            default_min_key_length=0,
        )

    def get_sorted_string(self, value: str) -> str:
        if self.sort_option.case_sensitive:
            return f"~{value}"
        return f"~{value.lower()}"


SorterRegistry.register_sorter(SortType.INTEGER.value, IntegerSorter)
SorterRegistry.register_sorter(SortType.STRING.value, StringSorter)
SorterRegistry.register_sorter(SortType.FLOAT.value, FloatSorter)
SorterRegistry.register_sorter(SortType.DATETIME.value, DateTimeSorter)
# SorterRegistry.register_sorter(SortType.CUSTOM.value, CustomSorterProxy)


class EnumSorter(CustomSorter):
    def __init__(
        self,
        sort_option: TsvFileSortOption,
        column_idx: int,
        column_name_indices: Dict[str, int],
        column_indices: Dict[int, str],
    ) -> None:
        super().__init__(
            sort_option,
            column_idx,
            column_name_indices,
            column_indices,
            default_min_key_length=20,
        )

        if "enum-class" in sort_option.custom_sorter_arguments:
            self.enum_class = sort_option.custom_sorter_arguments["enum-class"]
            try:
                if isinstance(self.enum_class, str):
                    splitted_value = self.enum_class.split(":")

                    if len(splitted_value) == 2:
                        self.enum_class = getattr(
                            sys.modules[splitted_value[0]], splitted_value[1]
                        )
                    else:
                        raise TsvSortException(
                            "Selected enum class name is not valid. Use format <module name>:<enum class name>"
                        )
            except Exception as exc:
                if isinstance(exc, TsvSortException):
                    raise exc
                raise TsvSortException(
                    "Selected class name is not converted to enum"
                ) from exc
        if not isinstance(self.enum_class, EnumMeta):
            raise TsvSortException("selected class is not enum")

        self.enum_string_map: Dict[str, str] = {}
        if "enum-value-map" in sort_option.custom_sorter_arguments:
            self.enum_value_map = sort_option.custom_sorter_arguments["enum-value-map"]
            if not isinstance(self.enum_value_map, dict) or not self.enum_value_map:
                raise TsvSortException("Enum map is not valid dict")

            for key in self.enum_value_map:
                value = self.enum_value_map[key]
                self.enum_string_map[str(key)] = str(value).zfill(self.min_key_length)

            enum_values = {str(e.value) for e in self.enum_class}
            missing_values = [x for x in enum_values if x not in self.enum_string_map]
            if missing_values:
                raise TsvSortException(
                    f"There are some missing values to map an enum {', '.join(missing_values)}"
                )

    def get_sorted_string(self, value: str) -> str:
        return self.enum_string_map[value]


SorterRegistry.register_custom_sorter("enum-sorter", EnumSorter)



from enum import Enum


class ColumnsStructure(str, Enum):
    SINGLE_COLUMN = "SINGLE_COLUMN"
    ONTOLOGY_COLUMN = "ONTOLOGY_COLUMN"
    SINGLE_COLUMN_AND_UNIT_ONTOLOGY = "SINGLE_COLUMN_AND_UNIT_ONTOLOGY"
    LINKED_COLUMN = "LINKED_COLUMN"
    ADDITIONAL_COLUMN = "ADDITIONAL_COLUMN"
    INVALID_MULTI_COLUMN = "INVALID_MULTI_COLUMN"


class IsaTableAdditionalColumn(str, Enum):
    UNIT = "Unit"
    TERM_SOURCE_REF = "Term Source REF"
    TERM_ACCSSION_NUMBER = "Term Accession Number"



import pathlib
import uuid
from typing import Dict, List

from metabolights_utils.tsv import model as actions
from metabolights_utils.tsv.actions.base import BaseTsvAction


class UpdateColumnsTsvAction(BaseTsvAction):
    def apply_action(
        self,
        source_file_path: pathlib.Path,
        target_file_path: pathlib.Path,
        action: actions.TsvUpdateColumnsAction,
        read_encoding: str = "utf-8",
        write_encoding: str = "utf-8",
    ) -> actions.TsvActionResult:
        result: actions.TsvActionResult = actions.TsvActionResult(action=action)
        if action.action_type != actions.TsvActionType.UPDATE_COLUMN_DATA:
            result.message = "Action name is not valid"
            return result

        action: actions.TsvUpdateColumnsAction = action

        columns: Dict[int, actions.TsvColumnData] = (
            action.columns if action.columns else {}
        )

        if not columns:
            result.message = "There is not target column index"
            return result

        column_indices: List[int] = list(columns.keys()).copy()
        column_indices.sort()

        if not action.id:
            uuid_value = str(uuid.uuid4().hex)
            action.id = uuid_value

        try:
            with open(source_file_path, "r", encoding=read_encoding) as source:
                header_line = source.readline()
                header_names = header_line.strip("\n").split("\t")
                selected_columns: List[int] = []
                for column_idx, value in enumerate(header_names):
                    if column_idx in column_indices:
                        if columns[column_idx].header_name != value:
                            result.message = (
                                f"Input header name does not match the actual one for index {column_idx}."
                                + f"Expected: {columns[column_idx].header_name}, found: {value}"
                            )
                            return result
                        selected_columns.append(column_idx)

                if len(selected_columns) != len(column_indices):
                    invalid_indices = [
                        x for x in column_indices if x not in selected_columns
                    ]
                    result.message = (
                        "Some column indices are not found :"
                        + f"{', '.join(invalid_indices)}"
                    )
                    return result

                invalid_targets = [
                    x for x in columns if x >= len(header_names) and x < 0
                ]
                if invalid_targets:
                    result.message = f"Target column indices are not valid: {', '.join(invalid_targets)}."
                    return result

                with open(target_file_path, "w", encoding=write_encoding) as target:
                    self.write_row(target, header_names)
                    row_index = 0
                    for line in source:
                        row = line.strip("\n").split("\t")
                        for column_idx, value in columns.items():
                            column_data: actions.TsvColumnData = value
                            if (
                                not column_data.values
                                or row_index in column_data.values
                            ):
                                row[column_idx] = column_data.values[row_index]

                        self.write_row(target, row)
                        row_index += 1

            result.success = True
        except Exception as exc:
            result.message = f"{str(exc)}"

        return result



from typing import List

from pydantic import Field
from typing_extensions import Annotated

from metabolights_utils.models.isa.common import (
    IsaTableFile,
    OntologyItem,
    OrganismAndOrganismPartPair,
)


class SamplesFile(IsaTableFile):
    sample_names: Annotated[
        List[str],
        Field(description="Unique values in `Sample Name` column in the sample file."),
    ] = []
    organisms: Annotated[
        List[OntologyItem],
        Field(description="Unique organism values in the sample file."),
    ] = []
    organism_parts: Annotated[
        List[OntologyItem],
        Field(description="Unique organism part values in the sample file."),
    ] = []
    organism_and_organism_part_pairs: Annotated[
        List[OrganismAndOrganismPartPair],
        Field(
            description="Unique organism, organism part, variant and sample type values in the sample file."
        ),
    ] = []
    variants: Annotated[
        List[OntologyItem],
        Field(description="Unique variant values in the sample file."),
    ] = []
    sample_types: Annotated[
        List[OntologyItem],
        Field(description="Unique sample type values in the sample file."),
    ] = []
